from datetime import timedelta

from celery import shared_task

from django.contrib.auth.models import User
from django.db import transaction, DatabaseError
from django.utils.timezone import now

from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from . import __title__
from .models import (
    DiscordWebhook,
    NotificationRule,
    ScheduledNotification,
    Timer,
)
from .utils import LoggerAddTag


logger = LoggerAddTag(get_extension_logger(__name__), __title__)
TASK_PRIO_HIGH = 4


@shared_task(base=QueueOnce, acks_late=True)
def send_messages_for_webhook(webhook_pk: int) -> None:
    """sends all currently queued messages for given webhook to Discord"""
    try:
        webhook = DiscordWebhook.objects.get(pk=webhook_pk)
    except DiscordWebhook.DoesNotExist:
        logger.error(
            "DiscordWebhook with pk = %s does not exist. Aborting.", webhook_pk
        )
    else:
        if not webhook.is_enabled:
            logger.info(
                "Tracker %s: DiscordWebhook disabled - skipping sending", webhook
            )
            return

        logger.info("Started sending messages to webhook %s", webhook)
        webhook.send_queued_messages()
        logger.info("Completed sending messages to webhook %s", webhook)


@shared_task(base=QueueOnce, bind=True, acks_late=True)
def send_scheduled_notification(self, scheduled_notification_pk: int) -> None:
    """Sends a scheduled notification for a timer based on a notification rule"""
    with transaction.atomic():
        try:
            scheduled_notification = (
                ScheduledNotification.objects.select_for_update().get(
                    pk=scheduled_notification_pk
                )
            )
        except (ScheduledNotification.DoesNotExist, DatabaseError):
            logger.info(
                "ScheduledNotification with pk = %s does not / no longer exist "
                "or is being processed by another task. Discarding.",
                scheduled_notification_pk,
                exc_info=True,
            )
        else:
            logger.debug(
                "send_scheduled_notification with task_id = %s, for %r",
                self.request.id,
                scheduled_notification,
            )
            timer = scheduled_notification.timer
            notification_rule = scheduled_notification.notification_rule
            if scheduled_notification.celery_task_id != self.request.id:
                logger.info(
                    "Discarding outdated scheduled notification: %r",
                    scheduled_notification,
                )
            elif not notification_rule.is_enabled:
                logger.info(
                    "Discarding scheduled notification based on disabled rule: %r",
                    scheduled_notification,
                )
            else:
                logger.info(
                    "Sending notifications for timer '%s' and rule '%s'",
                    timer,
                    notification_rule,
                )
                webhook = notification_rule.webhook
                if webhook.is_enabled:
                    minutes = round((timer.date - now()).total_seconds() / 60)
                    mod_text = "**important** " if timer.is_important else ""
                    content = (
                        f"The following {mod_text}structure timer will elapse "
                        f"in less than **{minutes:,}** minutes:"
                    )
                    timer.send_notification(
                        webhook=webhook,
                        content=notification_rule.prepend_ping_text(content),
                    )
                    send_messages_for_webhook.apply_async(
                        args=[webhook.pk], priority=TASK_PRIO_HIGH
                    )
                else:
                    logger.warning(
                        "Webhook not enabled for %r. Disgarding.",
                        scheduled_notification,
                    )

            scheduled_notification.delete()


@shared_task
def notify_about_new_timer(timer_pk: int, notification_rule_pk: int) -> None:
    try:
        timer = Timer.objects.get(pk=timer_pk)
        notification_rule = NotificationRule.objects.select_related("webhook").get(
            pk=notification_rule_pk
        )
    except Timer.DoesNotExist:
        logger.error("Timer with pk = %s does not exist. Aborting.", timer_pk)
    except NotificationRule.DoesNotExist:
        logger.error(
            "NotificationRule with pk = %s does not exist. Aborting.",
            notification_rule_pk,
        )
    else:
        if notification_rule.is_enabled and notification_rule.webhook.is_enabled:
            content = f"New timer added by **{timer.eve_character}**:"
            timer.send_notification(
                webhook=notification_rule.webhook,
                content=notification_rule.prepend_ping_text(content),
            )
            send_messages_for_webhook.apply_async(
                args=[notification_rule.webhook.pk], priority=TASK_PRIO_HIGH
            )


def _schedule_notification_for_timer(
    timer: Timer, notification_rule: NotificationRule
) -> ScheduledNotification:
    logger.info(
        "Scheduling fresh notification for timer #%d, rule #%d",
        timer.pk,
        notification_rule.pk,
    )
    notification_date = timer.date - timedelta(minutes=notification_rule.scheduled_time)
    scheduled_notification, _ = ScheduledNotification.objects.update_or_create(
        timer=timer,
        notification_rule=notification_rule,
        defaults={"timer_date": timer.date, "notification_date": notification_date},
    )
    result = send_scheduled_notification.apply_async(
        kwargs={"scheduled_notification_pk": scheduled_notification.pk},
        eta=timer.date - timedelta(minutes=notification_rule.scheduled_time),
        priority=TASK_PRIO_HIGH,
    )
    scheduled_notification.celery_task_id = result.task_id
    scheduled_notification.save()

    return scheduled_notification


def _revoke_notification_for_timer(
    scheduled_notification: ScheduledNotification,
) -> None:
    logger.info(
        "Removing stale notification for timer #%d, rule #%d",
        scheduled_notification.timer.pk,
        scheduled_notification.notification_rule.pk,
    )
    scheduled_notification.delete()


@shared_task(acks_late=True)
def schedule_notifications_for_timer(timer_pk: int, is_new: bool = False) -> None:
    """Schedules notifications for this timer based on notification rules"""
    try:
        timer = Timer.objects.get(pk=timer_pk)
    except Timer.DoesNotExist:
        logger.error("Timer with pk = %s does not exist. Aborting.", timer_pk)
    else:
        if timer.date > now():
            # trigger: newly created
            if is_new:
                rules = (
                    NotificationRule.objects.select_related("webhook")
                    .filter(
                        is_enabled=True,
                        trigger=NotificationRule.TRIGGER_NEW_TIMER_CREATED,
                        webhook__is_enabled=True,
                    )
                    .conforms_with_timer(timer)
                )
                if rules:
                    for rule in rules:
                        notify_about_new_timer.apply_async(
                            kwargs={
                                "timer_pk": timer.pk,
                                "notification_rule_pk": rule.pk,
                            },
                            priority=TASK_PRIO_HIGH,
                        )

            # trigger: timer elapses soon
            with transaction.atomic():
                # remove existing scheduled notifications if date has changed
                for obj in ScheduledNotification.objects.filter(timer=timer).exclude(
                    timer_date=timer.date
                ):
                    _revoke_notification_for_timer(scheduled_notification=obj)

                # schedule new notifications
                for notification_rule in NotificationRule.objects.filter(
                    is_enabled=True,
                    trigger=NotificationRule.TRIGGER_SCHEDULED_TIME_REACHED,
                ).conforms_with_timer(timer):
                    _schedule_notification_for_timer(
                        timer=timer, notification_rule=notification_rule
                    )
        else:
            logger.warning("Can not schedule notification for past timer: %s", timer)


@shared_task(acks_late=True)
def schedule_notifications_for_rule(notification_rule_pk: int) -> None:
    """Schedules notifications for all timers confirming with this rule.
    Will recreate all existing and still pending notifications
    """
    try:
        notification_rule = NotificationRule.objects.get(pk=notification_rule_pk)
    except NotificationRule.DoesNotExist:
        logger.error(
            "NotificationRule with pk = %s does not exist. Aborting.",
            notification_rule_pk,
        )
    else:
        if notification_rule.trigger == NotificationRule.TRIGGER_NEW_TIMER_CREATED:
            logger.error(
                "NotificationRule with pk = %s has the wrong trigger. Aborting.",
                notification_rule_pk,
            )
        else:
            logger.debug(
                "Checking scheduled notifications for: %s",
                notification_rule,
            )
            with transaction.atomic():
                for obj in notification_rule.schedulednotification_set.filter(
                    timer_date__gt=now()
                ):
                    _revoke_notification_for_timer(scheduled_notification=obj)

                for timer in Timer.objects.filter(
                    date__gt=now()
                ).conforms_with_notification_rule(notification_rule):
                    _schedule_notification_for_timer(
                        timer=timer, notification_rule=notification_rule
                    )


@shared_task
def send_test_message_to_webhook(webhook_pk: int, user_pk: int = None) -> None:
    """send a test message to given webhook.
    Optional inform user about result if user ok is given
    """
    try:
        webhook = DiscordWebhook.objects.get(pk=webhook_pk)
        if user_pk:
            user = User.objects.get(pk=user_pk)
        else:
            user = None
    except DiscordWebhook.DoesNotExist:
        logger.error(
            "DiscordWebhook with pk = %s does not exist. Aborting.", webhook_pk
        )
    except User.DoesNotExist:
        logger.warning("User with pk = %s does not exist. Aborting.", user_pk)
    else:
        logger.info("Sending test message to webhook %s", webhook)
        error_text, success = webhook.send_test_message(user)

        if user:
            message = (
                f"Error text: {error_text}\nCheck log files for details."
                if not success
                else "No errors"
            )
            level = "success" if success else "error"
            notify(
                user=user,
                title=(
                    f"{__title__}: Result of test message to webhook {webhook}: "
                    f"{level.upper()}"
                ),
                message=message,
                level=level,
            )


@shared_task
def housekeeping() -> None:
    """Perform housekeeping tasks"""
    logger.info("Performing housekeeping")
    deleted_count = Timer.objects.delete_obsolete()
    logger.info(f"Deleted {deleted_count:,} obsolete timers.")



from datetime import timedelta
from unittest.mock import patch
from io import StringIO

from django.core.management import call_command
from django.utils.timezone import now

from allianceauth.timerboard.models import Timer as AuthTimer
from eveuniverse.models import EveType

from . import LoadTestDataMixin, create_test_user
from ..models import Timer
from ..utils import NoSocketsTestCase, app_labels

PACKAGE_PATH = "structuretimers.management.commands"


@patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
@patch(PACKAGE_PATH + ".structuretimers_migrate_timers.get_input")
class TestMigirateTimers(LoadTestDataMixin, NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        if "timerboard" not in app_labels():
            raise KeyboardInterrupt(
                "App `timerboard` is not installed, which is required for this test"
            )

    def setUp(self) -> None:
        self.out = StringIO()
        self.user = create_test_user(self.character_1)
        self.auth_timer = AuthTimer.objects.create(
            system="Abune",
            planet_moon="Near Heydieles gate",
            structure="Astrahus",
            eve_time=now() + timedelta(hours=4),
            eve_character=self.character_1,
            eve_corp=self.corporation_1,
            user=self.user,
        )
        Timer.objects.all().delete()

    def test_full_armor_friendly(self, mock_get_input):
        mock_get_input.return_value = "Y"
        self.auth_timer.details = "Armor timer"
        self.auth_timer.objective = "Friendly"
        self.auth_timer.save()

        call_command("structuretimers_migrate_timers", stdout=self.out)

        new_timer = Timer.objects.first()
        self.assertEqual(new_timer.eve_solar_system, self.system_abune)
        self.assertEqual(new_timer.structure_type, self.type_astrahus)
        self.assertEqual(new_timer.timer_type, Timer.TYPE_ARMOR)
        self.assertEqual(new_timer.details_notes, "Armor timer")
        self.assertEqual(new_timer.objective, Timer.OBJECTIVE_FRIENDLY)
        self.assertEqual(new_timer.date, self.auth_timer.eve_time)
        self.assertEqual(new_timer.eve_character, self.character_1)
        self.assertEqual(new_timer.eve_corporation, self.corporation_1)
        self.assertEqual(new_timer.user, self.auth_timer.user)

    def test_hull_hostile(self, mock_get_input):
        mock_get_input.return_value = "Y"
        self.auth_timer.details = "Hull timer"
        self.auth_timer.objective = "Hostile"
        self.auth_timer.save()

        call_command("structuretimers_migrate_timers", stdout=self.out)

        new_timer = Timer.objects.first()
        self.assertEqual(new_timer.timer_type, Timer.TYPE_HULL)
        self.assertEqual(new_timer.objective, Timer.OBJECTIVE_HOSTILE)

    def test_anchoring(self, mock_get_input):
        mock_get_input.return_value = "Y"
        self.auth_timer.details = "Anchor timer"
        self.auth_timer.objective = "Neutral"
        self.auth_timer.save()

        call_command("structuretimers_migrate_timers", stdout=self.out)

        new_timer = Timer.objects.first()
        self.assertEqual(new_timer.timer_type, Timer.TYPE_ANCHORING)
        self.assertEqual(new_timer.objective, Timer.OBJECTIVE_NEUTRAL)

    def test_final_corp_timer(self, mock_get_input):
        mock_get_input.return_value = "Y"
        self.auth_timer.details = "Final timer"
        self.auth_timer.corp_timer = True
        self.auth_timer.save()

        call_command("structuretimers_migrate_timers", stdout=self.out)

        new_timer = Timer.objects.first()
        self.assertEqual(new_timer.timer_type, Timer.TYPE_FINAL)
        self.assertEqual(new_timer.visibility, Timer.VISIBILITY_CORPORATION)

    def test_moon_mining(self, mock_get_input):
        mock_get_input.return_value = "Y"
        self.auth_timer.structure = "Moon Mining Cycle"
        self.auth_timer.save()

        call_command("structuretimers_migrate_timers", stdout=self.out)

        new_timer = Timer.objects.first()
        self.assertEqual(new_timer.timer_type, Timer.TYPE_MOONMINING)
        self.assertEqual(new_timer.structure_type, EveType.objects.get(id=35835))

    def test_abort_on_unknown_solar_system(self, mock_get_input):
        mock_get_input.return_value = "Y"
        self.auth_timer.system = "Unknown"
        self.auth_timer.save()

        call_command("structuretimers_migrate_timers", stdout=self.out)

        self.assertFalse(Timer.objects.all().exists())

    def test_abort_on_unknown_structure_type(self, mock_get_input):
        mock_get_input.return_value = "Y"
        self.auth_timer.structure = "Unknown"
        self.auth_timer.save()

        call_command("structuretimers_migrate_timers", stdout=self.out)

        self.assertFalse(Timer.objects.all().exists())

    def test_do_not_create_duplicates(self, mock_get_input):
        mock_get_input.return_value = "Y"

        call_command("structuretimers_migrate_timers", stdout=self.out)
        call_command("structuretimers_migrate_timers", stdout=self.out)

        self.assertEqual(Timer.objects.all().count(), 1)



# Generated by Django 2.2.13 on 2020-08-15 19:17

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import multiselectfield.db.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveonline", "0012_index_additions"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("eveuniverse", "0002_load_eveunit"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app and see timers"),
                    ("create_timer", "Can create new timers and edit own timers"),
                    ("manage_timer", "Can edit and delete any timer"),
                    ("opsec_access", "Can create and see opsec timers"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="DiscordWebhook",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="short name to identify this webhook",
                        max_length=64,
                        unique=True,
                    ),
                ),
                (
                    "url",
                    models.CharField(
                        help_text="URL of this webhook, e.g. https://discordapp.com/api/webhooks/123456/abcdef",
                        max_length=255,
                        unique=True,
                    ),
                ),
                (
                    "notes",
                    models.TextField(
                        blank=True,
                        default=None,
                        help_text="you can add notes about this webhook here if you want",
                        null=True,
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        db_index=True,
                        default=True,
                        help_text="whether notifications are currently sent to this webhook",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="NotificationRule",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "trigger",
                    models.CharField(
                        choices=[
                            ("TC", "New timer created"),
                            ("TR", "Scheduled time reached"),
                        ],
                        help_text="Trigger for sending a notification",
                        max_length=2,
                    ),
                ),
                (
                    "scheduled_time",
                    models.PositiveIntegerField(
                        blank=True,
                        choices=[
                            (None, "---------"),
                            (0, "T - 0 minutes"),
                            (5, "T - 5 minutes"),
                            (10, "T - 10 minutes"),
                            (15, "T - 15 minutes"),
                            (30, "T - 30 minutes"),
                            (45, "T - 45 minutes"),
                            (60, "T - 60 minutes"),
                            (90, "T - 90 minutes"),
                            (120, "T - 120 minutes"),
                        ],
                        db_index=True,
                        default=None,
                        help_text="When to sent a notification in relation to when the timer elapses. Use together with `Scheduled time reached` trigger.",
                        null=True,
                    ),
                ),
                (
                    "ping_type",
                    models.CharField(
                        choices=[
                            ("PN", "(no ping)"),
                            ("PH", "@here"),
                            ("PE", "@everyone"),
                        ],
                        default="PN",
                        help_text="Options for pinging on notification",
                        max_length=2,
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        default=True, help_text="whether this rule is currently active"
                    ),
                ),
                (
                    "require_timer_types",
                    multiselectfield.db.fields.MultiSelectField(
                        blank=True,
                        choices=[
                            ("NO", "Unspecified"),
                            ("AR", "Armor"),
                            ("HL", "Hull"),
                            ("FI", "Final"),
                            ("AN", "Anchoring"),
                            ("UA", "Unanchoring"),
                            ("MM", "Moon Mining"),
                        ],
                        help_text="Timer must have one of the given timer types or leave blank to match any.",
                        max_length=20,
                    ),
                ),
                (
                    "exclude_timer_types",
                    multiselectfield.db.fields.MultiSelectField(
                        blank=True,
                        choices=[
                            ("NO", "Unspecified"),
                            ("AR", "Armor"),
                            ("HL", "Hull"),
                            ("FI", "Final"),
                            ("AN", "Anchoring"),
                            ("UA", "Unanchoring"),
                            ("MM", "Moon Mining"),
                        ],
                        help_text="Timer must NOT have one of the given timer types",
                        max_length=20,
                    ),
                ),
                (
                    "require_objectives",
                    multiselectfield.db.fields.MultiSelectField(
                        blank=True,
                        choices=[
                            ("UN", "undefined"),
                            ("HO", "hostile"),
                            ("FR", "friendly"),
                            ("NE", "neutral"),
                        ],
                        help_text="Timer must have one of the given objectives or leave blank to match any.",
                        max_length=11,
                    ),
                ),
                (
                    "exclude_objectives",
                    multiselectfield.db.fields.MultiSelectField(
                        blank=True,
                        choices=[
                            ("UN", "undefined"),
                            ("HO", "hostile"),
                            ("FR", "friendly"),
                            ("NE", "neutral"),
                        ],
                        help_text="Timer must NOT have one of the given objectives",
                        max_length=11,
                    ),
                ),
                (
                    "require_visibility",
                    multiselectfield.db.fields.MultiSelectField(
                        blank=True,
                        choices=[
                            ("UN", "unrestricted"),
                            ("AL", "Alliance only"),
                            ("CO", "Corporation only"),
                        ],
                        help_text="Visibility must be one of the selected or leave blank to match any.",
                        max_length=8,
                    ),
                ),
                (
                    "exclude_visibility",
                    multiselectfield.db.fields.MultiSelectField(
                        blank=True,
                        choices=[
                            ("UN", "unrestricted"),
                            ("AL", "Alliance only"),
                            ("CO", "Corporation only"),
                        ],
                        help_text="Visibility must NOT be one of the selected",
                        max_length=8,
                    ),
                ),
                (
                    "is_important",
                    models.CharField(
                        choices=[("AN", "any"), ("RQ", "required"), ("EX", "excluded")],
                        default="AN",
                        help_text="Wether the timer must be important",
                        max_length=2,
                    ),
                ),
                (
                    "is_opsec",
                    models.CharField(
                        choices=[("AN", "any"), ("RQ", "required"), ("EX", "excluded")],
                        default="AN",
                        help_text="Wether the timer must be OPSEC",
                        max_length=2,
                    ),
                ),
                (
                    "exclude_alliances",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Timer must NOT be created by one of the given alliances",
                        related_name="notification_rule_exclude_alliances",
                        to="eveonline.EveAllianceInfo",
                    ),
                ),
                (
                    "exclude_corporations",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Timer must NOT be created by one of the given corporations",
                        related_name="notification_rule_exclude_corporations",
                        to="eveonline.EveCorporationInfo",
                    ),
                ),
                (
                    "require_alliances",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Timer must be created by one of the given alliances or leave blank to match any.",
                        related_name="notification_rule_require_alliances",
                        to="eveonline.EveAllianceInfo",
                    ),
                ),
                (
                    "require_corporations",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Timer must be created by one of the given corporations or leave blank to match any.",
                        related_name="notification_rule_require_corporations",
                        to="eveonline.EveCorporationInfo",
                    ),
                ),
                (
                    "webhook",
                    models.ForeignKey(
                        help_text="The webhook all notifications are sent to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structuretimers.DiscordWebhook",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ScheduledNotification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timer_date", models.DateTimeField(db_index=True)),
                ("notification_date", models.DateTimeField(db_index=True)),
                ("celery_task_id", models.CharField(default="", max_length=765)),
                (
                    "notification_rule",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structuretimers.NotificationRule",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Timer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "timer_type",
                    models.CharField(
                        choices=[
                            ("NO", "Unspecified"),
                            ("AR", "Armor"),
                            ("HL", "Hull"),
                            ("FI", "Final"),
                            ("AN", "Anchoring"),
                            ("UA", "Unanchoring"),
                            ("MM", "Moon Mining"),
                        ],
                        default="NO",
                        max_length=2,
                    ),
                ),
                (
                    "location_details",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Additional information about the location of this structure, e.g. name of nearby planet / moon / gate",
                        max_length=254,
                    ),
                ),
                (
                    "structure_name",
                    models.CharField(blank=True, default="", max_length=254),
                ),
                (
                    "objective",
                    models.CharField(
                        choices=[
                            ("UN", "undefined"),
                            ("HO", "hostile"),
                            ("FR", "friendly"),
                            ("NE", "neutral"),
                        ],
                        default="UN",
                        max_length=2,
                    ),
                ),
                (
                    "date",
                    models.DateTimeField(
                        db_index=True, help_text="Date when this timer happens"
                    ),
                ),
                (
                    "is_important",
                    models.BooleanField(
                        default=False, help_text="Mark this timer as is_important"
                    ),
                ),
                (
                    "owner_name",
                    models.CharField(
                        blank=True,
                        default=None,
                        help_text="Name of the corporation owning the structure",
                        max_length=254,
                        null=True,
                    ),
                ),
                (
                    "is_opsec",
                    models.BooleanField(
                        db_index=True,
                        default=False,
                        help_text="Limit access to users with OPSEC clearance. Can be combined with visibility.",
                    ),
                ),
                (
                    "visibility",
                    models.CharField(
                        choices=[
                            ("UN", "unrestricted"),
                            ("AL", "Alliance only"),
                            ("CO", "Corporation only"),
                        ],
                        db_index=True,
                        default="UN",
                        help_text="The visibility of this timer can be limited to members of your organization",
                        max_length=2,
                    ),
                ),
                (
                    "details_image_url",
                    models.CharField(
                        blank=True,
                        default=None,
                        help_text="URL for details like a screenshot of the structure's fitting, e.g. https://www.example.com/route/image.jpg",
                        max_length=1024,
                        null=True,
                    ),
                ),
                (
                    "details_notes",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Notes with additional information about this timer",
                    ),
                ),
                (
                    "eve_alliance",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="Alliance of the user who created this timer",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="Timers",
                        to="eveonline.EveAllianceInfo",
                    ),
                ),
                (
                    "eve_character",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="Main character of the user who created this timer",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="Timers",
                        to="eveonline.EveCharacter",
                    ),
                ),
                (
                    "eve_corporation",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="Corporation of the user who created this timer",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="Timers",
                        to="eveonline.EveCorporationInfo",
                    ),
                ),
                (
                    "eve_solar_system",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.EveSolarSystem",
                    ),
                ),
                (
                    "notification_rules",
                    models.ManyToManyField(
                        help_text="Notification rules conforming with this timer",
                        through="structuretimers.ScheduledNotification",
                        to="structuretimers.NotificationRule",
                    ),
                ),
                (
                    "structure_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.EveType",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="Timers",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="schedulednotification",
            name="timer",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="structuretimers.Timer"
            ),
        ),
        migrations.AddConstraint(
            model_name="schedulednotification",
            constraint=models.UniqueConstraint(
                fields=("timer", "notification_rule"),
                name="unique_notification_schedule",
            ),
        ),
    ]



from datetime import timedelta
import json
from unittest.mock import Mock, patch

import requests

from django.contrib.auth.models import User
from django.http import HttpRequest
from django.test import TestCase
from django.utils import translation
from django.utils.html import mark_safe
from django.utils.timezone import now

from ..utils import (
    clean_setting,
    messages_plus,
    chunks,
    timeuntil_str,
    NoSocketsTestCase,
    SocketAccessError,
    app_labels,
    add_no_wrap_html,
    yesno_str,
    create_bs_button_html,
    create_bs_glyph_html,
    create_link_html,
    add_bs_label_html,
    get_site_base_url,
    JSONDateTimeDecoder,
    JSONDateTimeEncoder,
    generate_invalid_pk,
)
from ..utils import set_test_logger


MODULE_PATH = "structuretimers.utils"
logger = set_test_logger(MODULE_PATH, __file__)


class TestMessagePlus(TestCase):
    @patch(MODULE_PATH + ".messages", spec=True)
    def test_valid_call(self, mock_messages):
        messages_plus.debug(Mock(spec=HttpRequest), "Test Message")
        self.assertTrue(mock_messages.debug.called)
        call_args_list = mock_messages.debug.call_args_list
        args, kwargs = call_args_list[0]
        self.assertEqual(
            args[1],
            '<span class="glyphicon glyphicon-eye-open" '
            'aria-hidden="true"></span>&nbsp;&nbsp;'
            "Test Message",
        )

    def test_invalid_level(self):
        with self.assertRaises(ValueError):
            messages_plus._add_messages_icon(987, "Test Message")

    @patch(MODULE_PATH + ".messages")
    def test_all_levels(self, mock_messages):
        text = "Test Message"
        messages_plus.error(Mock(spec=HttpRequest), text)
        self.assertTrue(mock_messages.error.called)

        messages_plus.debug(Mock(spec=HttpRequest), text)
        self.assertTrue(mock_messages.debug.called)

        messages_plus.info(Mock(spec=HttpRequest), text)
        self.assertTrue(mock_messages.info.called)

        messages_plus.success(Mock(spec=HttpRequest), text)
        self.assertTrue(mock_messages.success.called)

        messages_plus.warning(Mock(spec=HttpRequest), text)
        self.assertTrue(mock_messages.warning.called)


class TestChunks(TestCase):
    def test_chunks(self):
        a0 = [1, 2, 3, 4, 5, 6]
        a1 = list(chunks(a0, 2))
        self.assertListEqual(a1, [[1, 2], [3, 4], [5, 6]])


class TestCleanSetting(TestCase):
    @patch(MODULE_PATH + ".settings")
    def test_default_if_not_set(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = Mock(spec=None)
        result = clean_setting(
            "TEST_SETTING_DUMMY",
            False,
        )
        self.assertEqual(result, False)

    @patch(MODULE_PATH + ".settings")
    def test_default_if_not_set_for_none(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = Mock(spec=None)
        result = clean_setting("TEST_SETTING_DUMMY", None, required_type=int)
        self.assertEqual(result, None)

    @patch(MODULE_PATH + ".settings")
    def test_true_stays_true(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = True
        result = clean_setting(
            "TEST_SETTING_DUMMY",
            False,
        )
        self.assertEqual(result, True)

    @patch(MODULE_PATH + ".settings")
    def test_false_stays_false(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = False
        result = clean_setting("TEST_SETTING_DUMMY", False)
        self.assertEqual(result, False)

    @patch(MODULE_PATH + ".settings")
    def test_default_for_invalid_type_bool(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = "invalid type"
        result = clean_setting("TEST_SETTING_DUMMY", False)
        self.assertEqual(result, False)

    @patch(MODULE_PATH + ".settings")
    def test_default_for_invalid_type_int(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = "invalid type"
        result = clean_setting("TEST_SETTING_DUMMY", 50)
        self.assertEqual(result, 50)

    @patch(MODULE_PATH + ".settings")
    def test_none_allowed_for_type_int(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = None
        result = clean_setting("TEST_SETTING_DUMMY", 50)
        self.assertIsNone(result)

    @patch(MODULE_PATH + ".settings")
    def test_default_if_below_minimum_1(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = -5
        result = clean_setting("TEST_SETTING_DUMMY", default_value=50)
        self.assertEqual(result, 50)

    @patch(MODULE_PATH + ".settings")
    def test_default_if_below_minimum_2(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = -50
        result = clean_setting("TEST_SETTING_DUMMY", default_value=50, min_value=-10)
        self.assertEqual(result, 50)

    @patch(MODULE_PATH + ".settings")
    def test_default_for_invalid_type_int_2(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = 1000
        result = clean_setting("TEST_SETTING_DUMMY", default_value=50, max_value=100)
        self.assertEqual(result, 50)

    @patch(MODULE_PATH + ".settings")
    def test_default_is_none_needs_required_type(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = "invalid type"
        with self.assertRaises(ValueError):
            clean_setting("TEST_SETTING_DUMMY", default_value=None)

    @patch(MODULE_PATH + ".settings")
    def test_when_value_in_choices_return_it(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = "bravo"
        result = clean_setting(
            "TEST_SETTING_DUMMY", default_value="alpha", choices=["alpha", "bravo"]
        )
        self.assertEqual(result, "bravo")

    @patch(MODULE_PATH + ".settings")
    def test_when_value_not_in_choices_return_default(self, mock_settings):
        mock_settings.TEST_SETTING_DUMMY = "charlie"
        result = clean_setting(
            "TEST_SETTING_DUMMY", default_value="alpha", choices=["alpha", "bravo"]
        )
        self.assertEqual(result, "alpha")


class TestTimeUntil(TestCase):
    def test_timeuntil(self):
        duration = timedelta(days=365 + 30 * 4 + 5, seconds=3600 * 14 + 60 * 33 + 10)
        expected = "1y 4mt 5d 14h 33m 10s"
        self.assertEqual(timeuntil_str(duration), expected)

        duration = timedelta(days=2, seconds=3600 * 14 + 60 * 33 + 10)
        expected = "2d 14h 33m 10s"
        self.assertEqual(timeuntil_str(duration), expected)

        duration = timedelta(days=2, seconds=3600 * 14 + 60 * 33 + 10)
        expected = "2d 14h 33m 10s"
        self.assertEqual(timeuntil_str(duration), expected)

        duration = timedelta(days=0, seconds=60 * 33 + 10)
        expected = "0h 33m 10s"
        self.assertEqual(timeuntil_str(duration), expected)

        duration = timedelta(days=0, seconds=10)
        expected = "0h 0m 10s"
        self.assertEqual(timeuntil_str(duration), expected)

        duration = timedelta(days=-10, seconds=-20)
        expected = ""
        self.assertEqual(timeuntil_str(duration), expected)


class TestNoSocketsTestCase(NoSocketsTestCase):
    def test_raises_exception_on_attempted_network_access(self):

        with self.assertRaises(SocketAccessError):
            requests.get("https://www.google.com")


class TestAppLabel(TestCase):
    def test_returns_set_of_app_labels(self):
        labels = app_labels()
        for label in ["authentication", "groupmanagement", "eveonline"]:
            self.assertIn(label, labels)


class TestHtmlHelper(TestCase):
    def test_add_no_wrap_html(self):
        expected = '<span style="white-space: nowrap;">Dummy</span>'
        self.assertEqual(add_no_wrap_html("Dummy"), expected)

    def test_yesno_str(self):
        with translation.override("en"):
            self.assertEqual(yesno_str(True), "yes")
            self.assertEqual(yesno_str(False), "no")
            self.assertEqual(yesno_str(None), "no")
            self.assertEqual(yesno_str(123), "no")
            self.assertEqual(yesno_str("xxxx"), "no")

    def test_add_bs_label_html(self):
        expected = '<div class="label label-danger">Dummy</div>'
        self.assertEqual(add_bs_label_html("Dummy", "danger"), expected)

    def test_create_link_html_default(self):
        expected = (
            '<a href="https://www.example.com" target="_blank">' "Example Link</a>"
        )
        self.assertEqual(
            create_link_html("https://www.example.com", "Example Link"), expected
        )

    def test_create_link_html(self):
        expected = '<a href="https://www.example.com">Example Link</a>'
        self.assertEqual(
            create_link_html("https://www.example.com", "Example Link", False), expected
        )
        expected = (
            '<a href="https://www.example.com">' "<strong>Example Link</strong></a>"
        )
        self.assertEqual(
            create_link_html(
                "https://www.example.com",
                mark_safe("<strong>Example Link</strong>"),
                False,
            ),
            expected,
        )

    def test_create_bs_glyph_html(self):
        expected = '<span class="glyphicon glyphicon-example"></span>'
        self.assertEqual(create_bs_glyph_html("example"), expected)

    def test_create_bs_button_html_default(self):
        expected = (
            '<a href="https://www.example.com" class="btn btn-info">'
            '<span class="glyphicon glyphicon-example"></span></a>'
        )
        self.assertEqual(
            create_bs_button_html("https://www.example.com", "example", "info"),
            expected,
        )

    def test_create_bs_button_html_disabled(self):
        expected = (
            '<a href="https://www.example.com" class="btn btn-info"'
            ' disabled="disabled">'
            '<span class="glyphicon glyphicon-example"></span></a>'
        )
        self.assertEqual(
            create_bs_button_html("https://www.example.com", "example", "info", True),
            expected,
        )


class TestGetSiteBaseUrl(NoSocketsTestCase):
    @patch(
        MODULE_PATH + ".settings.ESI_SSO_CALLBACK_URL",
        "https://www.mysite.com/sso/callback",
    )
    def test_return_url_if_url_defined_and_valid(self):
        expected = "https://www.mysite.com"
        self.assertEqual(get_site_base_url(), expected)

    @patch(
        MODULE_PATH + ".settings.ESI_SSO_CALLBACK_URL",
        "https://www.mysite.com/not-valid/",
    )
    def test_return_dummy_if_url_defined_but_not_valid(self):
        expected = ""
        self.assertEqual(get_site_base_url(), expected)

    @patch(MODULE_PATH + ".settings")
    def test_return_dummy_if_url_not_defined(self, mock_settings):
        delattr(mock_settings, "ESI_SSO_CALLBACK_URL")
        expected = ""
        self.assertEqual(get_site_base_url(), expected)


class TestJsonSerializer(NoSocketsTestCase):
    def test_encode_decode(self):
        my_dict = {"alpha": "hello", "bravo": now()}
        my_json = json.dumps(my_dict, cls=JSONDateTimeEncoder)
        my_dict_new = json.loads(my_json, cls=JSONDateTimeDecoder)
        self.assertDictEqual(my_dict, my_dict_new)


class TestGenerateInvalidPk(NoSocketsTestCase):
    def test_normal(self):
        User.objects.all().delete()
        User.objects.create(username="John Doe", password="dummy")
        invalid_pk = generate_invalid_pk(User)
        with self.assertRaises(User.DoesNotExist):
            User.objects.get(pk=invalid_pk)



from django.apps import AppConfig
from . import __version__


class StructureTimersConfig(AppConfig):
    name = "structuretimers"
    label = "structuretimers"
    verbose_name = f"Structure Timers v{__version__}"



from datetime import timedelta
import logging

from django.core.exceptions import PermissionDenied
from django.contrib.auth.decorators import login_required, permission_required
from django.http import JsonResponse, HttpResponseForbidden, HttpResponse
from django.utils.html import format_html, mark_safe
from django.shortcuts import render, reverse
from django.utils.translation import gettext_lazy as _
from django.utils.timezone import now

from django.views import View
from django.urls import reverse_lazy
from django.views.generic import CreateView, UpdateView, DeleteView
from django.contrib.auth.mixins import LoginRequiredMixin, PermissionRequiredMixin

from allianceauth.eveonline.evelinks import evewho, dotlan

from eveuniverse.models import EveSolarSystem, EveType

from . import __title__
from .app_settings import (
    STRUCTURETIMERS_DEFAULT_PAGE_LENGTH,
    STRUCTURETIMERS_PAGING_ENABLED,
)
from .forms import TimerForm
from .models import Timer
from .utils import (
    add_no_wrap_html,
    add_bs_label_html,
    create_fa_button_html,
    create_link_html,
    yesno_str,
    messages_plus,
)

logger = logging.getLogger(__name__)
DATETIME_FORMAT = "%Y-%m-%d %H:%M"
MAX_HOURS_PASSED = 2


@login_required
@permission_required("structuretimers.basic_access")
def timer_list(request):
    context = {
        "current_time": now().strftime("%Y-%m-%d %H:%M:%S"),
        "max_hours_expired": MAX_HOURS_PASSED,
        "title": __title__,
        "data_tables_page_length": STRUCTURETIMERS_DEFAULT_PAGE_LENGTH,
        "data_tables_paging": STRUCTURETIMERS_PAGING_ENABLED,
    }
    return render(request, "structuretimers/timer_list.html", context=context)


@login_required
@permission_required("structuretimers.basic_access")
def timer_list_data(request, tab_name):
    """returns timer list in JSON for AJAX call in timer_list view"""

    timers_qs = Timer.objects.all().visible_to_user(request.user)

    if tab_name == "current":
        timers_qs = timers_qs.filter(
            date__gte=now() - timedelta(hours=MAX_HOURS_PASSED)
        )
    else:
        timers_qs = timers_qs.filter(date__lt=now())

    timers_qs = timers_qs.select_related(
        "eve_solar_system__eve_constellation__eve_region",
        "structure_type",
        "eve_character",
        "eve_corporation",
        "eve_alliance",
    )
    data = list()
    for timer in timers_qs:
        # location
        location = create_link_html(
            dotlan.solar_system_url(timer.eve_solar_system.name),
            timer.eve_solar_system.name,
        )
        if timer.location_details:
            location += format_html("<br><em>{}</em>", timer.location_details)

        location += format_html(
            "<br>{}", timer.eve_solar_system.eve_constellation.eve_region.name
        )

        # structure & timer type & fitting image
        timer_type = add_bs_label_html(
            timer.get_timer_type_display(), timer.label_type_for_timer_type()
        )
        if timer.structure_type:
            structure_type_icon_url = timer.structure_type.icon_url(size=64)
            structure_type_name = timer.structure_type.name
        else:
            structure_type_icon_url = ""
            structure_type_name = "(unknown)"

        structure = format_html(
            '<div class="flex-container">'
            '  <div style="padding-top: 4px;"><img src="{}" width="40"></div>'
            '  <div style="text-align: left;">'
            "    {}&nbsp;<br>"
            "    {}"
            "  </div>"
            "</div>",
            structure_type_icon_url,
            mark_safe(add_bs_label_html(structure_type_name, "info")),
            timer_type,
        )

        # objective & tags
        tags = list()
        is_restricted = False
        if timer.is_opsec:
            tags.append(add_bs_label_html("OPSEC", "danger"))
            is_restricted = True

        if timer.visibility != Timer.VISIBILITY_UNRESTRICTED:
            tags.append(add_bs_label_html(timer.get_visibility_display(), "info"))
            is_restricted = True

        if timer.is_important:
            tags.append(add_bs_label_html("Important", "warning"))

        objective = format_html(
            "{}<br>{}",
            mark_safe(
                add_bs_label_html(
                    timer.get_objective_display(), timer.label_type_for_objective()
                )
            ),
            mark_safe(" ".join(tags)),
        )

        # name & owner
        if timer.owner_name:
            owner_name = timer.owner_name
            owner = owner_name
        else:
            owner = "-"
            owner_name = ""

        structure_name = timer.structure_name if timer.structure_name else "-"
        name = format_html("{}<br>{}", structure_name, owner)

        # creator
        if timer.eve_corporation:
            corporation_name = timer.eve_corporation.corporation_name
        else:
            corporation_name = "-"

        if timer.eve_character:
            creator = format_html(
                "{}<br>{}",
                create_link_html(
                    evewho.character_url(timer.eve_character.character_id),
                    timer.eve_character.character_name,
                ),
                corporation_name,
            )
        elif corporation_name:
            creator = corporation_name
        else:
            creator = "-"

        # visibility
        visibility = ""
        if timer.visibility == Timer.VISIBILITY_ALLIANCE and timer.eve_alliance:
            visibility = timer.eve_alliance.alliance_name
        elif timer.visibility == Timer.VISIBILITY_CORPORATION:
            visibility = corporation_name

        # actions
        actions = ""

        if timer.details_image_url or timer.details_notes:
            disabled_html = ""
            button_type = "primary"
            data_toggle = 'data-toggle="modal" data-target="#modalTimerDetails" '
            title = "Show details of this timer"
        else:
            button_type = "default"
            disabled_html = ' disabled="disabled"'
            data_toggle = ""
            title = "No details available"

        actions += (
            format_html(
                '<a type="button" id="timerboardBtnDetails" '
                f'class="btn btn-{button_type}" title="{title}"'
                f"{data_toggle}"
                f'data-timerpk="{timer.pk}"{disabled_html}>'
                '<i class="fas fa-search-plus"></i></a>'
            )
            + "&nbsp;"
        )

        if timer.user_can_edit(request.user):
            actions += (
                create_fa_button_html(
                    reverse("structuretimers:delete", args=(timer.pk,)),
                    "far fa-trash-alt",
                    "danger",
                    "Delete this timer",
                )
                + "&nbsp;"
                + create_fa_button_html(
                    reverse("structuretimers:edit", args=(timer.pk,)),
                    "far fa-edit",
                    "warning",
                    "Edit this timer",
                )
            )

        actions = add_no_wrap_html(actions)

        data.append(
            {
                "id": timer.id,
                "local_time": timer.date.isoformat(),
                "date": timer.date.isoformat(),
                "location": location,
                "structure_details": structure,
                "name_objective": name,
                "owner": objective,
                "creator": creator,
                "actions": actions,
                "timer_type_name": timer.get_timer_type_display(),
                "objective_name": timer.get_objective_display(),
                "system_name": timer.eve_solar_system.name,
                "region_name": timer.eve_solar_system.eve_constellation.eve_region.name,
                "structure_type_name": timer.structure_type.name,
                "owner_name": owner_name,
                "visibility": visibility,
                "opsec_str": yesno_str(timer.is_opsec),
                "is_opsec": timer.is_opsec,
                "is_passed": timer.date < now(),
                "is_important": timer.is_important,
                "is_restricted": is_restricted,
            }
        )
    return JsonResponse(data, safe=False)


@login_required
@permission_required("structuretimers.basic_access")
def get_timer_data(request, pk):
    """returns data for a timer"""
    timers_qs = Timer.objects.filter(pk=pk).visible_to_user(request.user)
    if timers_qs.exists():
        timer = timers_qs.first()
        data = {
            "display_name": str(timer),
            "structure_display_name": timer.structure_display_name,
            "date": timer.date.strftime(DATETIME_FORMAT),
            "details_image_url": timer.details_image_url
            if timer.details_image_url
            else "",
            "notes": timer.details_notes if timer.details_notes else "",
        }
        return JsonResponse(data, safe=False)
    else:
        return HttpResponseForbidden()


class BaseTimerView(LoginRequiredMixin, PermissionRequiredMixin, View):
    pass


class TimerManagementView(BaseTimerView):
    index_redirect = "structuretimers:timer_list"
    success_url = reverse_lazy(index_redirect)
    model = Timer
    form_class = TimerForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["title"] = "Edit Structure Timer"
        return context


class AddUpdateMixin:
    def get_form_kwargs(self):
        """
        Inject the request user into the kwargs passed to the form
        """
        kwargs = super().get_form_kwargs()
        kwargs.update({"user": self.request.user})
        return kwargs


class AddTimerView(TimerManagementView, AddUpdateMixin, CreateView):
    template_name_suffix = "_create_form"
    permission_required = (
        "structuretimers.basic_access",
        "structuretimers.create_timer",
    )

    def form_valid(self, form):
        result = super().form_valid(form)
        timer = self.object
        logger.info(
            "Created new timer in %s at %s by user %s",
            timer.eve_solar_system,
            timer.date,
            self.request.user,
        )
        messages_plus.success(
            self.request,
            _("Added new timer for %(type)s in %(system)s at %(time)s.")
            % {
                "type": timer.structure_type.name,
                "system": timer.eve_solar_system.name,
                "time": timer.date.strftime(DATETIME_FORMAT),
            },
        )
        return result


class EditTimerMixin:
    permission_required = "structuretimers.basic_access"

    def dispatch(self, request, *args, **kwargs) -> HttpResponse:
        response = super().dispatch(request, *args, **kwargs)
        if response.status_code == 200:
            if (
                not self.object.user_can_edit(self.request.user)
                or not Timer.objects.filter(pk=self.object.pk)
                .visible_to_user(self.request.user)
                .exists()
            ):
                raise PermissionDenied()

        return response


class EditTimerView(EditTimerMixin, TimerManagementView, AddUpdateMixin, UpdateView):
    template_name_suffix = "_update_form"

    """
    def form_valid(self, form):        
        timer = self.object        
        messages_plus.success(
            self.request, _('Saved changes to the timer: {}.').format(timer)
        )        
        return super().form_valid(form)
    """


class RemoveTimerView(EditTimerMixin, TimerManagementView, DeleteView):
    pass


@login_required
@permission_required("structuretimers.basic_access")
def select2_solar_systems(request):
    term = request.GET.get("term")
    if term:
        results = [
            {"id": row["id"], "text": row["name"]}
            for row in EveSolarSystem.objects.filter(name__istartswith=term).values(
                "id", "name"
            )
        ]
    else:
        results = None

    return JsonResponse({"results": results}, safe=False)


@login_required
@permission_required("structuretimers.basic_access")
def select2_structure_types(request):
    term = request.GET.get("term")
    if term:
        types_qs = (
            EveType.objects.filter(eve_group__eve_category_id__in=[65], published=True)
            | EveType.objects.filter(eve_group_id=365, published=True)
            | EveType.objects.filter(id=2233)
        )
        types_qs = (
            types_qs.select_related("eve_category", "eve_category__eve_group")
            .distinct()
            .filter(name__icontains=term)
        )
        results = [
            {"id": row["id"], "text": row["name"]}
            for row in types_qs.values("id", "name")
        ]
    else:
        results = None

    return JsonResponse({"results": results}, safe=False)



from unittest.mock import patch

from django_webtest import WebTest

from django.contrib.auth.models import User
from django.urls import reverse

from . import LoadTestDataMixin
from ..models import DiscordWebhook, NotificationRule, Timer


@patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
class TestNotificationRuleChangeList(LoadTestDataMixin, WebTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.webhook = DiscordWebhook.objects.create(
            name="Dummy", url="http://www.example.com"
        )
        cls.user = User.objects.create_superuser(
            "Bruce Wayne", "bruce@example.com", "password"
        )

    @patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
    def setUp(self) -> None:
        NotificationRule.objects.create(
            trigger=NotificationRule.TRIGGER_SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            webhook=self.webhook,
        )
        NotificationRule.objects.create(
            trigger=NotificationRule.TRIGGER_SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            require_timer_types=[Timer.TYPE_ARMOR],
            webhook=self.webhook,
        )
        rule = NotificationRule.objects.create(
            trigger=NotificationRule.TRIGGER_SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            webhook=self.webhook,
        )
        rule.require_corporations.add(self.corporation_1)
        NotificationRule.objects.create(
            trigger=NotificationRule.TRIGGER_SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            is_important=NotificationRule.CLAUSE_EXCLUDED,
            webhook=self.webhook,
        )

    def test_can_open_page_normally(self):
        # login
        self.app.set_user(self.user)

        # user tries to add new notification rule
        add_page = self.app.get(
            reverse("admin:structuretimers_notificationrule_changelist")
        )
        self.assertEqual(add_page.status_code, 200)


@patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
class TestNotificationRuleValidations(LoadTestDataMixin, WebTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.webhook = DiscordWebhook.objects.create(
            name="Dummy", url="http://www.example.com"
        )
        cls.user = User.objects.create_superuser(
            "Bruce Wayne", "bruce@example.com", "password"
        )
        cls.url_add = reverse("admin:structuretimers_notificationrule_add")
        cls.url_changelist = reverse(
            "admin:structuretimers_notificationrule_changelist"
        )

    def setUp(self) -> None:
        NotificationRule.objects.all().delete()

    def _open_page(self) -> object:
        # login
        self.app.set_user(self.user)

        # user tries to add new notification rule
        add_page = self.app.get(self.url_add)
        self.assertEqual(add_page.status_code, 200)
        form = add_page.form
        form["trigger"] = NotificationRule.TRIGGER_SCHEDULED_TIME_REACHED
        form["scheduled_time"] = NotificationRule.MINUTES_10
        form["webhook"] = self.webhook.pk
        return form

    def test_no_errors(self):
        form = self._open_page()
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, self.url_changelist)
        self.assertEqual(NotificationRule.objects.count(), 1)

    def test_can_not_have_same_options_timer_types(self):
        form = self._open_page()
        form["require_timer_types"] = [Timer.TYPE_ANCHORING, Timer.TYPE_HULL]
        form["exclude_timer_types"] = [Timer.TYPE_ANCHORING, Timer.TYPE_ARMOR]
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(NotificationRule.objects.count(), 0)

    def test_can_not_have_same_options_objectives(self):
        form = self._open_page()
        form["require_objectives"] = [Timer.OBJECTIVE_FRIENDLY, Timer.OBJECTIVE_HOSTILE]
        form["exclude_objectives"] = [Timer.OBJECTIVE_FRIENDLY, Timer.OBJECTIVE_NEUTRAL]
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(NotificationRule.objects.count(), 0)

    def test_can_not_have_same_options_visibility(self):
        form = self._open_page()
        form["require_visibility"] = [Timer.VISIBILITY_CORPORATION]
        form["exclude_visibility"] = [Timer.VISIBILITY_CORPORATION]
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(NotificationRule.objects.count(), 0)

    def test_can_not_have_same_options_corporations(self):
        form = self._open_page()
        form["require_corporations"] = [self.corporation_1.pk, self.corporation_3.pk]
        form["exclude_corporations"] = [self.corporation_1.pk]
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(NotificationRule.objects.count(), 0)

    def test_can_not_have_same_options_alliances(self):
        form = self._open_page()
        form["require_alliances"] = [self.alliance_1.pk, self.alliance_3.pk]
        form["exclude_alliances"] = [self.alliance_1.pk]
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(NotificationRule.objects.count(), 0)



import datetime
import logging
import imghdr

import requests

from django import forms
from django.utils import timezone
from django.utils.html import mark_safe
from django.core.validators import MaxValueValidator, MinValueValidator
from django.utils.translation import ugettext_lazy as _

from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo

from eveuniverse.models import EveSolarSystem, EveType

from .models import Timer

logger = logging.getLogger(__name__)


class TimerForm(forms.ModelForm):
    class Meta:
        model = Timer
        fields = (
            "eve_solar_system_2",
            "location_details",
            "structure_type_2",
            "timer_type",
            "structure_name",
            "owner_name",
            "objective",
            "days_left",
            "hours_left",
            "minutes_left",
            "details_image_url",
            "details_notes",
            "visibility",
            "is_opsec",
            "is_important",
        )

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user", None)
        if "instance" in kwargs and kwargs["instance"] is not None:
            # Do conversion from db datetime to days/hours/minutes
            # for appropriate fields
            my_instance = kwargs["instance"]
            current_time = timezone.now()
            td = my_instance.date - current_time
            initial = kwargs.pop("initial", dict())
            if "days_left" not in initial:
                initial.update({"days_left": td.days})
            if "hours_left" not in initial:
                initial.update({"hours_left": td.seconds // 3600})
            if "minutes_left" not in initial:
                initial.update({"minutes_left": td.seconds // 60 % 60})

            kwargs.update({"initial": initial})
            self.is_new = False
        else:
            my_instance = None
            self.is_new = True

        super().__init__(*args, **kwargs)

        if my_instance:
            self.fields["eve_solar_system_2"].widget.choices = [
                (
                    str(my_instance.eve_solar_system_id),
                    my_instance.eve_solar_system.name,
                )
            ]
            self.fields["structure_type_2"].widget.choices = [
                (
                    str(my_instance.structure_type_id),
                    my_instance.structure_type.name,
                )
            ]

    asterisk_html = '<i class="fas fa-asterisk"></i>'
    eve_solar_system_2 = forms.CharField(
        required=True,
        label=_(mark_safe(f"Solar System {asterisk_html}")),
        widget=forms.Select(attrs={"class": "select2-solar-systems"}),
    )
    structure_type_2 = forms.CharField(
        required=True,
        label=_(mark_safe(f"Structure Type {asterisk_html}")),
        widget=forms.Select(attrs={"class": "select2-structure-types"}),
    )
    objective = forms.ChoiceField(
        initial=Timer.OBJECTIVE_UNDEFINED,
        choices=Timer.OBJECTIVE_CHOICES,
        widget=forms.Select(attrs={"class": "select2-render"}),
    )
    timer_type = forms.ChoiceField(
        choices=Timer.TYPE_CHOICES,
        widget=forms.Select(attrs={"class": "select2-render"}),
    )
    visibility = forms.ChoiceField(
        choices=Timer.VISIBILITY_CHOICES,
        widget=forms.Select(attrs={"class": "select2-render"}),
    )
    days_left = forms.IntegerField(
        required=True,
        label=_(mark_safe(f"Days Remaining {asterisk_html}")),
        validators=[MinValueValidator(0)],
    )
    hours_left = forms.IntegerField(
        required=True,
        label=_(mark_safe(f"Hours Remaining {asterisk_html}")),
        validators=[MinValueValidator(0), MaxValueValidator(23)],
    )
    minutes_left = forms.IntegerField(
        required=True,
        label=_(mark_safe(f"Minutes Remaining {asterisk_html}")),
        validators=[MinValueValidator(0), MaxValueValidator(59)],
    )
    details_image_url = forms.URLField(required=False)

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data["eve_solar_system_2"]:
            try:
                solar_system = EveSolarSystem.objects.get(
                    id=cleaned_data["eve_solar_system_2"]
                )
            except EveSolarSystem.DoesNotExist:
                pass
            else:
                self.fields["eve_solar_system_2"].widget.choices = [
                    (
                        str(solar_system.id),
                        solar_system.name,
                    )
                ]

        if cleaned_data["structure_type_2"]:
            try:
                structure_type = EveType.objects.get(
                    id=cleaned_data["structure_type_2"]
                )
            except EveType.DoesNotExist:
                pass
            else:
                self.fields["structure_type_2"].widget.choices = [
                    (
                        str(structure_type.id),
                        structure_type.name,
                    )
                ]

        if cleaned_data["details_image_url"]:
            details_image_url = cleaned_data["details_image_url"]
            try:
                r = requests.get(details_image_url, stream=True, timeout=(3.0, 10.0))
                r.raise_for_status()
            except (
                requests.exceptions.ConnectionError,
                requests.exceptions.HTTPError,
            ):
                logger.warning(
                    f"Failed to load image from URL: {details_image_url}", exc_info=True
                )
                raise forms.ValidationError(
                    {
                        "details_image_url": _(
                            "Failed to load image file. Please double check URL."
                        )
                    },
                    code="details_url_failed_to_load",
                )
            else:
                image_type = imghdr.what(None, h=r.content)
                if image_type not in {"gif", "jpeg", "png"}:
                    logger.warning(
                        f"{image_type} is not a valid image type "
                        "for URL: {details_image_url}"
                    )
                    raise forms.ValidationError(
                        {
                            "details_image_url": _(
                                _(
                                    "URL does not point to a valid image file. "
                                    "Valid types are: gif, jpeg, png"
                                )
                            )
                        },
                        code="details_url_unsupported_type",
                    )

    def save(self, commit=True):
        timer = super(TimerForm, self).save(commit=False)

        # character / corporation / alliance
        if self.is_new:
            character = self.user.profile.main_character
            try:
                alliance = character.alliance
            except EveAllianceInfo.DoesNotExist:
                alliance = EveAllianceInfo.objects.create_alliance(
                    character.alliance_id
                )

            try:
                corporation = character.corporation
            except EveCorporationInfo.DoesNotExist:
                corporation = EveCorporationInfo.objects.create_corporation(
                    character.corporation_id
                )

            logger.debug(
                "Determined timer save request is on behalf of character %s corporation %s",
                character,
                corporation,
            )
            timer.eve_character = character
            timer.eve_corporation = corporation
            timer.eve_alliance = alliance
            timer.user = self.user

        # calculate future time
        future_time = datetime.timedelta(
            days=self.cleaned_data["days_left"],
            hours=self.cleaned_data["hours_left"],
            minutes=self.cleaned_data["minutes_left"],
        )
        current_time = timezone.now()
        date = current_time + future_time
        logger.debug(
            "Determined timer eve time is %s - current time %s, adding %s",
            date,
            current_time,
            future_time,
        )
        timer.date = date

        # structure type
        timer.structure_type_id = self.cleaned_data["structure_type_2"]
        timer.eve_solar_system_id = self.cleaned_data["eve_solar_system_2"]

        if commit:
            timer.save()
        return timer



from .utils import clean_setting

# Will not schedule notifications for timers
# which have elapsed more than x minutes ago
STRUCTURETIMERS_MAX_AGE_FOR_NOTIFICATIONS = clean_setting(
    "STRUCTURETIMERS_MAX_AGE_FOR_NOTIFICATIONS", 60
)

# Wether notifications for timers are scheduled at all
STRUCTURETIMERS_NOTIFICATIONS_ENABLED = clean_setting(
    "STRUCTURETIMERS_NOTIFICATIONS_ENABLED", True
)

# Minimum age in days for a timer to be considered obsolete
# Obsolete timers will automatically be deleted
STRUCTURETIMERS_TIMERS_OBSOLETE_AFTER_DAYS = clean_setting(
    "STRUCTURETIMERS_TIMERS_OBSOLETE_AFTER_DAYS", default_value=30, min_value=1
)

# Default page size for timerboard.
# Must be an integer value from the current options as seen in the app.
STRUCTURETIMERS_DEFAULT_PAGE_LENGTH = clean_setting(
    "STRUCTURETIMERS_DEFAULT_PAGE_LENGTH", 10
)

# Wether paging is enabled on the timerboard
STRUCTURETIMERS_PAGING_ENABLED = clean_setting("STRUCTURETIMERS_PAGING_ENABLED", True)



from django.db import models
from django.core.management.base import BaseCommand
from django.utils.timezone import now

from eveuniverse.models import EveSolarSystem

from ...models import Timer
from ...utils import app_labels


def get_input(text):
    """wrapped input to enable unit testing / patching"""
    return input(text)


# objectives
OBJECTIVE_CHOICES = [
    (Timer.OBJECTIVE_HOSTILE, "Hostile"),
    (Timer.OBJECTIVE_FRIENDLY, "Friendly"),
    (Timer.OBJECTIVE_NEUTRAL, "Neutral"),
    (Timer.OBJECTIVE_UNDEFINED, "Undefined"),
]

# structure types
STRUCTURE_TYPE_POCO = 2233
STRUCTURE_TYPE_I_HUB = 2233
STRUCTURE_TYPE_TCU = 32226
STRUCTURE_TYPE_POS_S = 20062
STRUCTURE_TYPE_POS_M = 20061
STRUCTURE_TYPE_POS_L = 16213
STRUCTURE_TYPE_ASTRAHUS = 35832
STRUCTURE_TYPE_FORTIZAR = 35833
STRUCTURE_TYPE_KEEPSTAR = 35834
STRUCTURE_TYPE_RAITARU = 35825
STRUCTURE_TYPE_AZBEL = 35826
STRUCTURE_TYPE_SOTIYO = 35827
STRUCTURE_TYPE_ATHANOR = 35835
STRUCTURE_TYPE_TATARA = 35836
STRUCTURE_TYPE_PHAROLUX = 35840
STRUCTURE_TYPE_TENEBREX = 37534
STRUCTURE_TYPE_ANSIBLEX = 35841
STRUCTURE_TYPE_STATION = 1529

STRUCTURE_CHOICES = [
    (STRUCTURE_TYPE_POCO, "POCO"),
    (STRUCTURE_TYPE_I_HUB, "I-HUB"),
    (STRUCTURE_TYPE_TCU, "TCU"),
    (STRUCTURE_TYPE_POS_S, "POS[S]"),
    (STRUCTURE_TYPE_POS_M, "POS[M]"),
    (STRUCTURE_TYPE_POS_L, "POS[L]"),
    (STRUCTURE_TYPE_ASTRAHUS, "Astrahus"),
    (STRUCTURE_TYPE_FORTIZAR, "Fortizar"),
    (STRUCTURE_TYPE_KEEPSTAR, "Keepstar"),
    (STRUCTURE_TYPE_RAITARU, "Raitaru"),
    (STRUCTURE_TYPE_AZBEL, "Azbel"),
    (STRUCTURE_TYPE_SOTIYO, "Sotiyo"),
    (STRUCTURE_TYPE_ATHANOR, "Athanor"),
    (STRUCTURE_TYPE_TATARA, "Tatara"),
    (STRUCTURE_TYPE_PHAROLUX, "Pharolux Cyno Beacon"),
    (STRUCTURE_TYPE_TENEBREX, "Tenebrex Cyno Jammer"),
    (STRUCTURE_TYPE_ANSIBLEX, "Ansiblex Jump Gate"),
    (STRUCTURE_TYPE_ATHANOR, "Moon Mining Cycle"),
]


class Command(BaseCommand):
    help = (
        "Removes all app-related data from the database. "
        "Run this command before zero migrations, "
        "which would otherwise fail due to FK constraints."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "-t",
            "--test",
            action="store_true",
            help="Perform's a test run. Does not actually migrate, but will show potential issues.",
        )

    def _migrate_timers(
        self, auth_timers_qs: models.QuerySet, is_test: bool = False
    ) -> None:
        objective_map = {x[1]: x[0] for x in OBJECTIVE_CHOICES}
        structure_map = {x[1]: x[0] for x in STRUCTURE_CHOICES}
        migrated_count = 0
        skipped_count = 0
        for auth_timer in auth_timers_qs:
            try:
                eve_solar_system = EveSolarSystem.objects.get(name=auth_timer.system)
            except EveSolarSystem.DoesNotExist:
                self.stdout.write(
                    self.style.WARNING(
                        f"Can not migrate timer '{auth_timer}', "
                        f"due to invalid solar system: {auth_timer.system}"
                    )
                )
                skipped_count += 1
                continue

            if auth_timer.structure in structure_map:
                structure_type_id = structure_map[auth_timer.structure]
            else:
                self.stdout.write(
                    self.style.WARNING(
                        f"Can not migrate timer '{auth_timer}', "
                        f"due to invalid structure type: {auth_timer.structure}"
                    )
                )
                skipped_count += 1
                continue

            if auth_timer.objective in objective_map:
                objective = objective_map[auth_timer.objective]
            else:
                objective = Timer.OBJECTIVE_UNDEFINED

            if auth_timer.corp_timer:
                visibility = Timer.VISIBILITY_CORPORATION
            else:
                visibility = Timer.VISIBILITY_UNRESTRICTED

            details_lower = auth_timer.details.lower()
            if "Moon Mining Cycle" in auth_timer.structure:
                timer_type = Timer.TYPE_MOONMINING
            elif "armor" in details_lower:
                timer_type = Timer.TYPE_ARMOR
            elif "hull" in details_lower:
                timer_type = Timer.TYPE_HULL
            elif "final" in details_lower:
                timer_type = Timer.TYPE_FINAL
            elif "unanchor" in details_lower:
                timer_type = Timer.TYPE_UNANCHORING
            elif "anchor" in details_lower:
                timer_type = Timer.TYPE_ANCHORING
            else:
                timer_type = Timer.TYPE_NONE

            if not is_test:
                try:
                    Timer.objects.get(
                        eve_solar_system=eve_solar_system,
                        structure_type_id=structure_type_id,
                        date=auth_timer.eve_time,
                        details_notes=auth_timer.details,
                    )
                    self.stdout.write(
                        self.style.WARNING(
                            f"Skipping timer '{auth_timer}', since it already exists"
                        )
                    )
                    skipped_count += 1

                except Timer.DoesNotExist:
                    Timer.objects.create(
                        eve_solar_system=eve_solar_system,
                        structure_type_id=structure_type_id,
                        date=auth_timer.eve_time,
                        user=auth_timer.user,
                        location_details=auth_timer.planet_moon,
                        timer_type=timer_type,
                        details_notes=auth_timer.details,
                        objective=objective,
                        is_important=auth_timer.important,
                        visibility=visibility,
                        eve_character=auth_timer.eve_character,
                        eve_corporation=auth_timer.eve_corp,
                        eve_alliance=auth_timer.eve_corp.alliance,
                    )
                    migrated_count += 1

        self.stdout.write(
            f"Results: Migrated: {migrated_count} - Skipped: {skipped_count} "
            f"- Total: {auth_timers_qs.count()}"
        )

    def handle(self, *args, **options):
        if "timerboard" not in app_labels():
            self.stdout.write(
                self.style.WARNING(
                    "The timerboard app from Auth is not installed. Aborted."
                )
            )
        else:
            from allianceauth.timerboard.models import Timer as AuthTimer

            auth_timers_qs = AuthTimer.objects.filter(eve_time__gt=now())
            self.stdout.write(
                f"This command will migrate {auth_timers_qs.count()} "
                "pending timers from the "
                "Auth's timerboard app to this app. "
            )
            self.stdout.write(
                "Note that timers that have been migrated previously "
                "will not be migrated again "
                "unless the migrated timers have been changed significantly."
            )
            is_test = options["test"]
            if is_test:
                self.stdout.write("Test run - will not migrate")

            user_input = get_input("Are you sure you want to proceed? (Y/n)?")
            if user_input == "Y":
                self.stdout.write("Starting migrating timers. Please stand by.")
                self._migrate_timers(auth_timers_qs, is_test=is_test)
                self.stdout.write(self.style.SUCCESS("Migration complete!"))
            else:
                self.stdout.write(self.style.WARNING("Aborted."))



from django.test import TestCase

from eveuniverse.tools.testdata import create_testdata, ModelSpec

from . import test_data_filename


class CreateEveUniverseTestData(TestCase):
    def test_create_testdata(self):
        testdata_spec = {
            "EveType": ModelSpec(
                ids=[35832, 35825, 35835],
                include_children=False,
            ),
            "EveSolarSystem": ModelSpec(
                ids=[30004984, 30045339],
                include_children=False,
            ),
        }
        create_testdata(testdata_spec, test_data_filename())



from datetime import timedelta
import json
from unittest.mock import patch

from django.contrib.auth.models import User
from django.test import TestCase, RequestFactory
from django.urls import reverse
from django.utils.timezone import now

from ..models import Timer
from .. import views
from . import LoadTestDataMixin, create_test_user, add_permission_to_user_by_name


MODULE_PATH = "structures.views"


def get_json_response(response: object):
    return json.loads(response.content.decode("utf-8"))


@patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
class TestViewBase(LoadTestDataMixin, TestCase):
    @patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
    def setUp(self):
        self.factory = RequestFactory()

        # user
        self.user_1 = create_test_user(self.character_1)
        self.user_2 = create_test_user(self.character_2)
        self.user_2 = add_permission_to_user_by_name(
            "structuretimers.manage_timer", self.user_2
        )
        self.user_3 = create_test_user(self.character_3)

        # timers
        self.timer_1 = Timer(
            structure_name="Timer 1",
            location_details="Near the star",
            date=now() + timedelta(hours=4),
            eve_character=self.character_1,
            eve_corporation=self.corporation_1,
            user=self.user_1,
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            owner_name="Big Boss",
            details_image_url="http://www.example.com/dummy.png",
            details_notes="Some notes",
        )
        self.timer_1.save()
        self.timer_2 = Timer(
            structure_name="Timer 2",
            date=now() - timedelta(hours=8),
            eve_character=self.character_1,
            eve_corporation=self.corporation_1,
            user=self.user_1,
            eve_solar_system=self.system_abune,
            structure_type=self.type_raitaru,
            is_important=True,
        )
        self.timer_2.save()
        self.timer_3 = Timer(
            structure_name="Timer 3",
            date=now() - timedelta(hours=8),
            eve_character=self.character_1,
            eve_corporation=self.corporation_1,
            user=self.user_1,
            eve_solar_system=self.system_enaluri,
            structure_type=self.type_astrahus,
        )
        self.timer_3.save()


class TestListData(TestViewBase):
    def _send_request(self, tab_name: str = "current", user: User = None) -> set:
        if not user:
            user = self.user_1

        request = self.factory.get(
            reverse("structuretimers:timer_list_data", args=[tab_name])
        )
        request.user = user
        response = views.timer_list_data(request, tab_name)
        self.assertEqual(response.status_code, 200)
        return {x["id"] for x in get_json_response(response)}

    def test_timer_list_view_loads(self):
        request = self.factory.get(reverse("structuretimers:timer_list"))
        request.user = self.user_1
        response = views.timer_list(request)
        self.assertEqual(response.status_code, 200)

    def test_timer_list_data_current_and_past(self):
        # test current timers
        timer_ids = self._send_request("current")
        expected = {self.timer_1.id}
        self.assertSetEqual(timer_ids, expected)

        # test past timers
        timer_ids = self._send_request("past")
        expected = {self.timer_2.id, self.timer_3.id}
        self.assertSetEqual(timer_ids, expected)

    def test_show_corp_restricted_to_corp_member(self):
        timer_4 = Timer(
            structure_name="Timer 4",
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            date=now() + timedelta(hours=8),
            eve_character=self.character_1,
            eve_corporation=self.corporation_1,
            user=self.user_2,
            visibility=Timer.VISIBILITY_CORPORATION,
        )
        timer_4.save()
        timer_ids = self._send_request()
        expected = {self.timer_1.id, timer_4.id}
        self.assertSetEqual(timer_ids, expected)

    def test_dont_show_corp_restricted_to_non_corp_member(self):
        timer = Timer(
            structure_name="Timer 4",
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            date=now() + timedelta(hours=8),
            eve_character=self.character_3,
            eve_corporation=self.corporation_3,
            user=self.user_3,
            visibility=Timer.VISIBILITY_CORPORATION,
        )
        timer.save()
        timer_ids = self._send_request()
        expected = {self.timer_1.id}
        self.assertSetEqual(timer_ids, expected)

    def test_show_alliance_restricted_to_alliance_member(self):
        timer_4 = Timer(
            structure_name="Timer 4",
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            date=now() + timedelta(hours=8),
            eve_character=self.character_1,
            eve_corporation=self.corporation_1,
            eve_alliance=self.alliance_1,
            user=self.user_2,
            visibility=Timer.VISIBILITY_ALLIANCE,
        )
        timer_4.save()
        timer_ids = self._send_request()
        expected = {self.timer_1.id, timer_4.id}
        self.assertSetEqual(timer_ids, expected)

    def test_dont_show_alliance_restricted_to_non_alliance_member(self):
        timer = Timer(
            structure_name="Timer 4",
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            date=now() + timedelta(hours=8),
            eve_character=self.character_3,
            eve_corporation=self.corporation_3,
            eve_alliance=self.alliance_3,
            user=self.user_3,
            visibility=Timer.VISIBILITY_ALLIANCE,
        )
        timer.save()
        timer_ids = self._send_request()
        expected = {self.timer_1.id}
        self.assertSetEqual(timer_ids, expected)

    def test_show_opsec_restricted_to_opsec_member(self):
        self.user_1 = add_permission_to_user_by_name(
            "structuretimers.opsec_access", self.user_1
        )
        timer_4 = Timer(
            structure_name="Timer 4",
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            date=now() + timedelta(hours=8),
            eve_character=self.character_3,
            eve_corporation=self.corporation_3,
            user=self.user_3,
            is_opsec=True,
        )
        timer_4.save()
        timer_ids = self._send_request()
        expected = {self.timer_1.id, timer_4.id}
        self.assertSetEqual(timer_ids, expected)

    def test_dont_show_opsec_restricted_to_non_opsec_member(self):
        timer = Timer(
            structure_name="Timer 4",
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            date=now() + timedelta(hours=8),
            eve_character=self.character_3,
            eve_corporation=self.corporation_3,
            user=self.user_3,
            is_opsec=True,
        )
        timer.save()
        timer_ids = self._send_request()
        expected = {self.timer_1.id}
        self.assertSetEqual(timer_ids, expected)

    def test_dont_show_opsec_corp_restricted_to_opsec_member_other_corp(self):
        self.user_1 = add_permission_to_user_by_name(
            "structuretimers.opsec_access", self.user_1
        )
        timer = Timer(
            structure_name="Timer 4",
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            date=now() + timedelta(hours=8),
            eve_character=self.character_3,
            eve_corporation=self.corporation_3,
            user=self.user_3,
            is_opsec=True,
            visibility=Timer.VISIBILITY_CORPORATION,
        )
        timer.save()
        timer_ids = self._send_request()
        expected = {self.timer_1.id}
        self.assertSetEqual(timer_ids, expected)

    def test_show_corp_timer_to_creator_of_different_corp(self):
        timer_4 = Timer.objects.create(
            structure_name="Timer 4",
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            date=now() + timedelta(hours=8),
            eve_character=self.character_3,
            eve_corporation=self.corporation_3,
            visibility=Timer.VISIBILITY_CORPORATION,
            user=self.user_1,
        )
        timer_ids = self._send_request()
        expected = {self.timer_1.id, timer_4.id}
        self.assertSetEqual(timer_ids, expected)

    def test_show_alliance_timer_to_creator_of_different_alliance(self):
        timer_4 = Timer.objects.create(
            structure_name="Timer 4",
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            date=now() + timedelta(hours=8),
            eve_character=self.character_3,
            eve_alliance=self.alliance_3,
            eve_corporation=self.corporation_3,
            visibility=Timer.VISIBILITY_ALLIANCE,
            user=self.user_1,
        )
        timer_ids = self._send_request()
        expected = {self.timer_1.id, timer_4.id}
        self.assertSetEqual(timer_ids, expected)

    def test_can_show_timers_without_user_character_corporation_or_alliance(self):
        timer_4 = Timer.objects.create(
            structure_name="Timer 4",
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            date=now() + timedelta(hours=8),
        )
        timer_ids = self._send_request()
        expected = {self.timer_1.id, timer_4.id}
        self.assertSetEqual(timer_ids, expected)

    def test_list_for_manager(self):
        timer_ids = self._send_request(user=self.user_2)
        expected = {self.timer_1.id}
        self.assertSetEqual(timer_ids, expected)


class TestGetTimerData(TestViewBase):
    def test_normal(self):
        request = self.factory.get(
            reverse("structuretimers:get_timer_data", args=[self.timer_1.pk])
        )
        request.user = self.user_1
        response = views.get_timer_data(request, self.timer_1.pk)
        self.assertEqual(response.status_code, 200)

    def test_forbidden(self):
        self.timer_1.is_opsec = True
        self.timer_1.save()
        request = self.factory.get(
            reverse("structuretimers:get_timer_data", args=[self.timer_1.pk])
        )
        request.user = self.user_3
        response = views.get_timer_data(request, self.timer_1.pk)
        self.assertEqual(response.status_code, 403)


class TestSelect2Views(LoadTestDataMixin, TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.user_1 = create_test_user(cls.character_1)

    def test_select2_solar_systems_normal(self):
        request = self.factory.get(
            reverse("structuretimers:select2_solar_systems"), data={"term": "abu"}
        )
        request.user = self.user_1
        response = views.select2_solar_systems(request)
        self.assertEqual(response.status_code, 200)
        data = get_json_response(response)
        self.assertEqual(data, {"results": [{"id": 30004984, "text": "Abune"}]})

    def test_select2_solar_systems_empty(self):
        request = self.factory.get(reverse("structuretimers:select2_solar_systems"))
        request.user = self.user_1
        response = views.select2_solar_systems(request)
        self.assertEqual(response.status_code, 200)
        data = get_json_response(response)
        self.assertEqual(data, {"results": None})

    def test_select2_structure_types_normal(self):
        request = self.factory.get(
            reverse("structuretimers:select2_structure_types"), data={"term": "ast"}
        )
        request.user = self.user_1
        response = views.select2_structure_types(request)
        self.assertEqual(response.status_code, 200)
        data = get_json_response(response)
        self.assertEqual(data, {"results": [{"id": 35832, "text": "Astrahus"}]})

    def test_select2_structure_types_empty(self):
        request = self.factory.get(reverse("structuretimers:select2_structure_types"))
        request.user = self.user_1
        response = views.select2_structure_types(request)
        self.assertEqual(response.status_code, 200)
        data = get_json_response(response)
        self.assertEqual(data, {"results": None})



"""Utility functions and classes for tests"""

from django.contrib.auth.models import User

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import (
    EveCharacter,
    EveCorporationInfo,
    EveAllianceInfo,
)
from allianceauth.tests.auth_utils import AuthUtils

from eveuniverse.models import EveSolarSystem, EveType

from .testdata.load_eveuniverse import load_eveuniverse


def add_main_to_user(user: User, character: EveCharacter):
    CharacterOwnership.objects.create(
        user=user, owner_hash="x1" + character.character_name, character=character
    )
    user.profile.main_character = character
    user.profile.save()


def create_test_user(character: EveCharacter) -> User:
    User.objects.filter(username=character.character_name).delete()
    user = AuthUtils.create_user(character.character_name)
    add_main_to_user(user, character)
    AuthUtils.add_permission_to_user_by_name("structuretimers.basic_access", user)
    user = User.objects.get(pk=user.pk)
    return user


def add_permission_to_user_by_name(
    perm: str, user: User, disconnect_signals=False
) -> User:
    """adds permission to given user by name and returns updated user object"""
    AuthUtils.add_permission_to_user_by_name(perm, user)
    return User.objects.get(pk=user.pk)


class LoadTestDataMixin:
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        load_eveuniverse()
        cls.type_astrahus = EveType.objects.get(id=35832)
        cls.type_raitaru = EveType.objects.get(id=35825)
        cls.system_abune = EveSolarSystem.objects.get(id=30004984)
        cls.system_enaluri = EveSolarSystem.objects.get(id=30045339)

        EveCharacter.objects.filter(character_id__in=[1001, 1002, 1003]).delete()
        EveCorporationInfo.objects.filter(corporation_id__in=[2001, 2003]).delete()
        EveAllianceInfo.objects.filter(alliance_id__in=[3001, 3003]).delete()
        cls.character_1 = EveCharacter.objects.create(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )
        cls.corporation_1 = EveCorporationInfo.objects.create(
            corporation_id=cls.character_1.corporation_id,
            corporation_name=cls.character_1.corporation_name,
            member_count=99,
        )
        cls.alliance_1 = EveAllianceInfo.objects.create(
            alliance_id=cls.character_1.alliance_id,
            alliance_name=cls.character_1.alliance_name,
            executor_corp_id=cls.corporation_1.corporation_id,
        )
        cls.character_2 = EveCharacter.objects.create(
            character_id=1002,
            character_name="Clark Kent",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )
        cls.character_3 = EveCharacter.objects.create(
            character_id=1003,
            character_name="Lex Luthor",
            corporation_id=2003,
            corporation_name="Lex Corp",
            alliance_id=3003,
            alliance_name="Lex Holding",
        )
        cls.corporation_3 = EveCorporationInfo.objects.create(
            corporation_id=cls.character_3.corporation_id,
            corporation_name=cls.character_3.corporation_name,
            member_count=666,
        )
        cls.alliance_3 = EveAllianceInfo.objects.create(
            alliance_id=cls.character_3.alliance_id,
            alliance_name=cls.character_3.alliance_name,
            executor_corp_id=cls.corporation_3.corporation_id,
        )



import logging
from django.core.management import call_command
from django.core.management.base import BaseCommand

from ... import __title__
from ...utils import LoggerAddTag


logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class Command(BaseCommand):
    help = "Preloads data required for this app from ESI"

    def handle(self, *args, **options):
        call_command(
            "eveuniverse_load_types",
            __title__,
            "--category_id",
            "65",
            "--group_id",
            "365",
            "--type_id",
            "2233",
        )



from datetime import timedelta
from unittest.mock import patch, Mock

from celery import Task

from django.test import TestCase, TransactionTestCase
from django.utils.timezone import now

from eveuniverse.models import EveSolarSystem, EveType

from . import LoadTestDataMixin
from ..models import DiscordWebhook, NotificationRule, ScheduledNotification, Timer
from ..tasks import (
    send_messages_for_webhook,
    schedule_notifications_for_timer,
    schedule_notifications_for_rule,
    send_scheduled_notification,
    notify_about_new_timer,
)
from .testdata.load_eveuniverse import load_eveuniverse
from ..utils import generate_invalid_pk

MODULE_PATH = "structuretimers.tasks"


class TestCaseBase(LoadTestDataMixin, TestCase):
    @patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
    def setUp(self) -> None:
        self.webhook = DiscordWebhook.objects.create(
            name="Dummy", url="http://www.example.com"
        )
        self.webhook.clear_queue()
        self.rule = NotificationRule.objects.create(
            trigger=NotificationRule.TRIGGER_SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_15,
            webhook=self.webhook,
        )
        self.timer = Timer.objects.create(
            structure_name="Test_1",
            eve_solar_system=self.system_abune,
            structure_type=self.type_raitaru,
            date=now() + timedelta(minutes=30),
        )


@patch(MODULE_PATH + ".DiscordWebhook.send_queued_messages", spec=True)
@patch(MODULE_PATH + ".logger", spec=True)
class TestSendMessagesForWebhook(TestCaseBase):
    def test_normal(self, mock_logger, mock_send_queued_messages):
        send_messages_for_webhook(self.webhook.pk)
        self.assertEqual(mock_send_queued_messages.call_count, 1)
        self.assertEqual(mock_logger.info.call_count, 2)
        self.assertEqual(mock_logger.error.call_count, 0)

    def test_invalid_pk(self, mock_logger, mock_send_queued_messages):
        send_messages_for_webhook(generate_invalid_pk(DiscordWebhook))
        self.assertEqual(mock_send_queued_messages.call_count, 0)
        self.assertEqual(mock_logger.info.call_count, 0)
        self.assertEqual(mock_logger.error.call_count, 1)

    def test_disabled_webhook(self, mock_logger, mock_send_queued_messages):
        self.webhook.is_enabled = False
        self.webhook.save()

        send_messages_for_webhook(self.webhook.pk)
        self.assertEqual(mock_send_queued_messages.call_count, 0)
        self.assertEqual(mock_logger.info.call_count, 1)
        self.assertEqual(mock_logger.error.call_count, 0)


@patch(MODULE_PATH + ".notify_about_new_timer", spec=True)
@patch(MODULE_PATH + ".send_scheduled_notification", spec=True)
class TestScheduleNotificationForTimer(TestCaseBase):
    def test_normal(self, mock_send_notification, mock_send_notification_for_timer):
        """
        given no notifications scheduled
        when called for timer with matching notification rule
        then schedules new notification
        """
        mock_send_notification.apply_async.return_value.task_id = "my_task_id"

        schedule_notifications_for_timer(timer_pk=self.timer.pk, is_new=True)

        self.assertTrue(mock_send_notification.apply_async.called)
        self.assertTrue(
            ScheduledNotification.objects.filter(
                timer=self.timer, notification_rule=self.rule
            )
        )

    def test_remove_old_notifications(
        self, mock_send_notification, mock_send_notification_for_timer
    ):
        """
        given existing notification
        when called for timer with matching notification rule and changed date
        then deletes existing notification and schedules new notification
        """
        mock_send_notification.apply_async.return_value.task_id = "my_task_id"
        notification_old = ScheduledNotification.objects.create(
            timer=self.timer,
            notification_rule=self.rule,
            timer_date=self.timer.date + timedelta(minutes=5),
            notification_date=self.timer.date - timedelta(minutes=5),
            celery_task_id="99",
        )

        schedule_notifications_for_timer(timer_pk=self.timer.pk, is_new=True)

        self.assertTrue(mock_send_notification.apply_async.called)
        self.assertTrue(
            ScheduledNotification.objects.filter(
                timer=self.timer, notification_rule=self.rule
            )
        )
        self.assertFalse(
            ScheduledNotification.objects.filter(pk=notification_old.pk).exists()
        )

    def test_notification_for_new_timer(
        self, mock_send_notification, mock_send_notification_for_timer
    ):
        """
        given notification rule for sending new timers exists
        when called for timer
        then send new notification
        """
        self.rule.is_enabled = False
        self.rule.save()
        rule = NotificationRule.objects.create(
            trigger=NotificationRule.TRIGGER_NEW_TIMER_CREATED, webhook=self.webhook
        )
        schedule_notifications_for_timer(timer_pk=self.timer.pk, is_new=True)

        self.assertTrue(mock_send_notification_for_timer.apply_async.called)
        _, kwargs = mock_send_notification_for_timer.apply_async.call_args
        self.assertEqual(kwargs["kwargs"]["timer_pk"], self.timer.pk)
        self.assertEqual(kwargs["kwargs"]["notification_rule_pk"], rule.pk)

    def test_no_notification_for_new_timer_if_no_rule(
        self, mock_send_notification, mock_send_notification_for_timer
    ):
        """
        given no notification rule for sending new timers exists
        when called for timer
        then no notification is sent
        """
        self.rule.is_enabled = False
        self.rule.save()
        schedule_notifications_for_timer(timer_pk=self.timer.pk, is_new=True)

        self.assertFalse(mock_send_notification_for_timer.apply_async.called)


@patch(MODULE_PATH + ".send_scheduled_notification", spec=True)
class TestScheduleNotificationForRule(TestCaseBase):
    def test_normal(self, mock_send_notification):
        """
        given no notifications scheduled
        when called for notification rule with matching timer
        then schedules new notification
        """
        mock_send_notification.apply_async.return_value.task_id = "my_task_id"

        schedule_notifications_for_rule(self.rule.pk)

        self.assertTrue(mock_send_notification.apply_async.called)
        self.assertTrue(
            ScheduledNotification.objects.filter(
                timer=self.timer, notification_rule=self.rule
            )
        )

    def test_remove_old_notifications(self, mock_send_notification):
        """
        given existing notification
        when called for notification rule with matching timer
        then deletes existing notification and schedules new notification
        """
        mock_send_notification.apply_async.return_value.task_id = "my_task_id"
        notification_old = ScheduledNotification.objects.create(
            timer=self.timer,
            notification_rule=self.rule,
            timer_date=self.timer.date + timedelta(minutes=5),
            notification_date=self.timer.date - timedelta(minutes=5),
            celery_task_id="99",
        )

        schedule_notifications_for_rule(self.rule.pk)

        self.assertTrue(mock_send_notification.apply_async.called)
        self.assertTrue(
            ScheduledNotification.objects.filter(
                timer=self.timer, notification_rule=self.rule
            )
        )
        self.assertFalse(
            ScheduledNotification.objects.filter(pk=notification_old.pk).exists()
        )


@patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
@patch(MODULE_PATH + ".send_messages_for_webhook", spec=True)
class TestSendScheduledNotification(TransactionTestCase):
    @patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
    def setUp(self) -> None:
        load_eveuniverse()
        self.type_raitaru = EveType.objects.get(id=35825)
        self.system_abune = EveSolarSystem.objects.get(id=30004984)
        self.webhook = DiscordWebhook.objects.create(
            name="Dummy", url="http://www.example.com"
        )
        self.webhook.clear_queue()
        self.rule = NotificationRule.objects.create(
            trigger=NotificationRule.TRIGGER_SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_15,
            webhook=self.webhook,
        )
        self.timer = Timer.objects.create(
            structure_name="Test_1",
            eve_solar_system=self.system_abune,
            structure_type=self.type_raitaru,
            date=now() + timedelta(minutes=30),
        )
        ScheduledNotification.objects.all().delete()

    def test_normal(self, mock_send_messages_for_webhook):
        """
        when this notification is correctly scheduled
        then send the notification
        """
        scheduled_notification = ScheduledNotification.objects.create(
            timer=self.timer,
            notification_rule=self.rule,
            celery_task_id="my-id-123",
            timer_date=now() + timedelta(hours=1),
            notification_date=now() + timedelta(minutes=30),
        )
        mock_task = Mock(spec=Task)
        mock_task.request.id = "my-id-123"
        send_scheduled_notification_inner = (
            send_scheduled_notification.__wrapped__.__func__
        )
        send_scheduled_notification_inner(
            mock_task, scheduled_notification_pk=scheduled_notification.pk
        )
        self.assertTrue(mock_send_messages_for_webhook.apply_async.called)

    def test_revoked_notification(self, mock_send_messages_for_webhook):
        """
        when this is not the right task instance
        then discard this notification
        """
        scheduled_notification = ScheduledNotification.objects.create(
            timer=self.timer,
            notification_rule=self.rule,
            celery_task_id="my-id-123",
            timer_date=now() + timedelta(hours=1),
            notification_date=now() + timedelta(minutes=30),
        )
        mock_task = Mock(**{"request.id": "my-id-456"})
        send_scheduled_notification_inner = (
            send_scheduled_notification.__wrapped__.__func__
        )
        send_scheduled_notification_inner(
            mock_task, scheduled_notification_pk=scheduled_notification.pk
        )
        self.assertFalse(mock_send_messages_for_webhook.apply_async.called)

    def test_rule_disabled(self, mock_send_messages_for_webhook):
        """
        when the notification rule for this scheduled notification is disabled
        then discard notification
        """
        self.rule.is_enabled = False
        self.rule.save()
        scheduled_notification = ScheduledNotification.objects.create(
            timer=self.timer,
            notification_rule=self.rule,
            celery_task_id="my-id-123",
            timer_date=now() + timedelta(hours=1),
            notification_date=now() + timedelta(minutes=30),
        )
        mock_task = Mock(spec=Task)
        mock_task.request.id = "my-id-123"
        send_scheduled_notification_inner = (
            send_scheduled_notification.__wrapped__.__func__
        )
        send_scheduled_notification_inner(
            mock_task, scheduled_notification_pk=scheduled_notification.pk
        )
        self.assertFalse(mock_send_messages_for_webhook.apply_async.called)


@patch(MODULE_PATH + ".send_messages_for_webhook", spec=True)
class TestSendNotificationForTimer(TestCaseBase):
    def setUp(self) -> None:
        super().setUp()
        self.rule_2 = NotificationRule.objects.create(
            trigger=NotificationRule.TRIGGER_NEW_TIMER_CREATED, webhook=self.webhook
        )

    def test_normal(self, mock_send_messages_for_webhook):
        """
        given rule for notifying about new timers exist
        when calling task for a timer
        then send notification for that timer
        """
        notify_about_new_timer(self.timer.pk, self.rule_2.pk)

        self.assertTrue(mock_send_messages_for_webhook.apply_async.called)
        _, kwargs = mock_send_messages_for_webhook.apply_async.call_args
        self.assertListEqual(kwargs["args"], [self.webhook.pk])

    def test_webhook_disabled(self, mock_send_messages_for_webhook):
        """
        given rule for notifying about new timers does NOT exist
        when calling task for a timer
        then send notification for that timer
        """
        notify_about_new_timer(self.timer.pk, self.rule_2.pk)

        self.assertTrue(mock_send_messages_for_webhook.apply_async.called)
        _, kwargs = mock_send_messages_for_webhook.apply_async.call_args
        self.assertListEqual(kwargs["args"], [self.webhook.pk])



from django.conf.urls import url

from . import views

app_name = "structuretimers"

urlpatterns = [
    url(r"^$", views.timer_list, name="timer_list"),
    url(r"^add/$", views.AddTimerView.as_view(), name="add"),
    url(r"^remove/(?P<pk>\w+)$", views.RemoveTimerView.as_view(), name="delete"),
    url(r"^edit/(?P<pk>\w+)$", views.EditTimerView.as_view(), name="edit"),
    url(
        r"^list_data/(?P<tab_name>\w+)$", views.timer_list_data, name="timer_list_data"
    ),
    url(r"^get_timer_data/(?P<pk>\w+)$", views.get_timer_data, name="get_timer_data"),
    url(
        r"^select2_solar_systems/$",
        views.select2_solar_systems,
        name="select2_solar_systems",
    ),
    url(
        r"^select2_structure_types/$",
        views.select2_structure_types,
        name="select2_structure_types",
    ),
]



from datetime import timedelta
from unittest.mock import patch

from requests.exceptions import ConnectionError as NewConnectionError, HTTPError

from allianceauth.tests.auth_utils import AuthUtils
from django_webtest import WebTest

from django.urls import reverse
from django.utils.timezone import now
from django.test import TestCase
from django.test.utils import override_settings

from . import LoadTestDataMixin, create_test_user, add_permission_to_user_by_name
from ..models import DiscordWebhook, Timer
from ..tasks import send_test_message_to_webhook
from .testdata import test_image_filename, test_data_filename


def bytes_from_file(filename, chunksize=8192):
    with open(filename, "rb") as f:
        while True:
            chunk = f.read(chunksize)
            if chunk:
                for b in chunk:
                    yield b
            else:
                break


@patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
class TestUI(LoadTestDataMixin, WebTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # user 1
        cls.user_1 = create_test_user(cls.character_1)

        # user 2
        cls.user_2 = create_test_user(cls.character_2)
        cls.user_2 = add_permission_to_user_by_name(
            "structuretimers.create_timer", cls.user_2
        )

    @patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
    def setUp(self) -> None:
        self.timer_1 = Timer.objects.create(
            structure_name="Timer 1",
            date=now() + timedelta(hours=4),
            eve_character=self.character_2,
            eve_corporation=self.corporation_1,
            user=self.user_2,
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
        )
        self.timer_2 = Timer.objects.create(
            structure_name="Timer 2",
            date=now() - timedelta(hours=8),
            eve_character=self.character_2,
            eve_corporation=self.corporation_1,
            user=self.user_2,
            eve_solar_system=self.system_abune,
            structure_type=self.type_raitaru,
        )
        self.timer_3 = Timer.objects.create(
            structure_name="Timer 3",
            date=now() - timedelta(hours=8),
            eve_character=self.character_2,
            eve_corporation=self.corporation_1,
            user=self.user_2,
            eve_solar_system=self.system_enaluri,
            structure_type=self.type_astrahus,
        )
        self.timer_3.save()

    def test_add_new_timer(self):
        """
        when user has permissions
        then he can create a new timer
        """

        # login
        self.app.set_user(self.user_2)

        # user opens timerboard
        timerboard = self.app.get(reverse("structuretimers:timer_list"))
        self.assertEqual(timerboard.status_code, 200)

        # user clicks on "Add Timer"
        add_timer = timerboard.click(href=reverse("structuretimers:add"))
        self.assertEqual(add_timer.status_code, 200)

        # user enters data and clicks create
        form = add_timer.forms["add-timer-form"]
        form["days_left"] = 1
        form["structure_name"] = "Timer 4"
        form["eve_solar_system_2"].force_value([str(self.system_abune.id)])
        form["structure_type_2"].force_value([str(self.type_astrahus.id)])
        form["days_left"] = 1
        form["hours_left"] = 2
        form["minutes_left"] = 3
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structuretimers:timer_list"))
        self.assertTrue(Timer.objects.filter(structure_name="Timer 4").exists())

    def test_add_new_timer_without_permission(self):
        """
        given a user does not have permissions
        when trying to access page for adding new timers
        then he is redirected back to dashboard
        """

        # login
        self.app.set_user(self.user_1)

        # user opens timerboard
        timerboard = self.app.get(reverse("structuretimers:timer_list"))
        self.assertEqual(timerboard.status_code, 200)

        # Add button not shown to user
        with self.assertRaises(IndexError):
            timerboard.click(href=reverse("structuretimers:add"))

        # direct request redirects user back to dashboard
        response = self.app.get(reverse("structuretimers:add"))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("authentication:dashboard"))

    def test_add_new_timer_validation_1(self):
        """
        when user entered invalid day
        then page can not be submitted and error is shown to user
        """

        # login
        self.app.set_user(self.user_2)

        # user clicks on "Add Timer"
        add_timer = self.app.get(reverse("structuretimers:add"))
        self.assertEqual(add_timer.status_code, 200)

        # user enters data and clicks create
        form = add_timer.forms["add-timer-form"]
        form["days_left"] = 1
        form["structure_name"] = "Timer 4"
        form["eve_solar_system_2"].force_value([str(self.system_abune.id)])
        form["structure_type_2"].force_value([str(self.type_astrahus.id)])
        form["days_left"] = -1
        form["hours_left"] = 2
        form["minutes_left"] = 3
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("correct the input errors", response.text)

    @patch("structuretimers.forms.requests.get", spec=True)
    def test_add_new_timer_validation_2(self, mock_get):
        """
        when user entered invalid day
        then page can not be submitted and error is shown to user
        """
        mock_get.side_effect = NewConnectionError

        # login
        self.app.set_user(self.user_2)

        # user clicks on "Add Timer"
        add_timer = self.app.get(reverse("structuretimers:add"))
        self.assertEqual(add_timer.status_code, 200)

        # user enters data and clicks create
        form = add_timer.forms["add-timer-form"]
        form["days_left"] = 1
        form["structure_name"] = "Timer 4"
        form["eve_solar_system_2"].force_value([str(self.system_abune.id)])
        form["structure_type_2"].force_value([str(self.type_astrahus.id)])
        form["days_left"] = 1
        form["hours_left"] = 2
        form["minutes_left"] = 3
        form["details_image_url"] = "http://www.example.com/image.png"
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("correct the input errors", response.text)

    @patch("structuretimers.forms.requests.get", spec=True)
    def test_add_new_timer_validation_3(self, mock_get):
        """
        when user entered invalid day
        then page can not be submitted and error is shown to user
        """
        mock_get.side_effect = HTTPError

        # login
        self.app.set_user(self.user_2)

        # user clicks on "Add Timer"
        add_timer = self.app.get(reverse("structuretimers:add"))
        self.assertEqual(add_timer.status_code, 200)

        # user enters data and clicks create
        form = add_timer.forms["add-timer-form"]
        form["days_left"] = 1
        form["structure_name"] = "Timer 4"
        form["eve_solar_system_2"].force_value([str(self.system_abune.id)])
        form["structure_type_2"].force_value([str(self.type_astrahus.id)])
        form["days_left"] = 1
        form["hours_left"] = 2
        form["minutes_left"] = 3
        form["details_image_url"] = "http://www.example.com/image.png"
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("correct the input errors", response.text)

    @patch("structuretimers.forms.requests.get", spec=True)
    def test_add_new_timer_validation_4(self, mock_get):
        """
        when user provides valid detail image link
        then timer is created
        """
        image_file = bytearray(bytes_from_file(test_image_filename()))
        mock_get.return_value.content = image_file

        # login
        self.app.set_user(self.user_2)

        # user clicks on "Add Timer"
        add_timer = self.app.get(reverse("structuretimers:add"))
        self.assertEqual(add_timer.status_code, 200)

        # user enters data and clicks create
        form = add_timer.forms["add-timer-form"]
        form["days_left"] = 1
        form["structure_name"] = "Timer 4"
        form["eve_solar_system_2"].force_value([str(self.system_abune.id)])
        form["structure_type_2"].force_value([str(self.type_astrahus.id)])
        form["days_left"] = 1
        form["hours_left"] = 2
        form["minutes_left"] = 3
        form["details_image_url"] = "http://www.example.com/image.png"
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structuretimers:timer_list"))
        self.assertTrue(Timer.objects.filter(structure_name="Timer 4").exists())

    @patch("structuretimers.forms.requests.get", spec=True)
    def test_add_new_timer_validation_5(self, mock_get):
        """
        when user provides invalid file link
        then page can not be submitted and error is shown to user
        """
        image_file = bytearray(bytes_from_file(test_data_filename()))
        mock_get.return_value.content = image_file

        # login
        self.app.set_user(self.user_2)

        # user clicks on "Add Timer"
        add_timer = self.app.get(reverse("structuretimers:add"))
        self.assertEqual(add_timer.status_code, 200)

        # user enters data and clicks create
        form = add_timer.forms["add-timer-form"]
        form["days_left"] = 1
        form["structure_name"] = "Timer 4"
        form["eve_solar_system_2"].force_value([str(self.system_abune.id)])
        form["structure_type_2"].force_value([str(self.type_astrahus.id)])
        form["days_left"] = 1
        form["hours_left"] = 2
        form["minutes_left"] = 3
        form["details_image_url"] = "http://www.example.com/image.png"
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("correct the input errors", response.text)

    def test_edit_existing_timer(self):
        """
        when user has permissions
        then he can edit an existing timer
        """

        # login
        self.app.set_user(self.user_2)

        # user opens timerboard
        timerboard = self.app.get(reverse("structuretimers:timer_list"))
        self.assertEqual(timerboard.status_code, 200)

        # user clicks on "Edit Timer" for timer 1
        edit_timer = self.app.get(
            reverse("structuretimers:edit", args=[self.timer_1.pk])
        )
        self.assertEqual(edit_timer.status_code, 200)

        # user enters data and clicks create
        form = edit_timer.forms["add-timer-form"]
        form["owner_name"] = "The Boys"
        response = form.submit()
        self.timer_1.refresh_from_db()

        # assert results
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structuretimers:timer_list"))
        self.assertEqual(self.timer_1.owner_name, "The Boys")

    def test_edit_timer_without_permission_1(self):
        """
        given a user does not have permissions
        when trying to access page for timer edit
        then he is redirected back to dashboard
        """

        # login
        self.app.set_user(self.user_1)

        # user tries to access page for edit directly
        response = self.app.get(reverse("structuretimers:edit", args=[self.timer_1.pk]))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("authentication:dashboard"))

    def test_edit_timer_of_others_without_permission_2(self):
        """
        given a user has permission to create tiemrs
        when trying to access page for timer edit of another user
        then he is redirected back to dashboard
        """

        # login
        user_3 = create_test_user(self.character_3)
        user_3 = add_permission_to_user_by_name("structuretimers.create_timer", user_3)
        self.app.set_user(user_3)

        # user tries to access page for edit directly
        response = self.app.get(reverse("structuretimers:edit", args=[self.timer_1.pk]))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("authentication:dashboard"))

    def test_edit_timer_of_others_with_manager_permission(self):
        """
        when a user has manager permission
        then he can edit timers of others
        """

        # login
        user_3 = create_test_user(self.character_3)
        user_3 = add_permission_to_user_by_name("structuretimers.manage_timer", user_3)
        self.app.set_user(user_3)

        # user opens timerboard
        timerboard = self.app.get(reverse("structuretimers:timer_list"))
        self.assertEqual(timerboard.status_code, 200)

        # user clicks on "Edit Timer" for timer 1
        edit_timer = self.app.get(
            reverse("structuretimers:edit", args=[self.timer_1.pk])
        )
        self.assertEqual(edit_timer.status_code, 200)

        # user enters data and clicks create
        form = edit_timer.forms["add-timer-form"]
        form["owner_name"] = "The Boys"
        response = form.submit()
        self.timer_1.refresh_from_db()

        # assert results
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structuretimers:timer_list"))
        self.assertEqual(self.timer_1.owner_name, "The Boys")

    def test_manager_tries_to_edit_corp_restricted_timer_of_others(self):
        """
        given a user has permission to create and manage timers
        when trying to access page for timer edit of a corp restricted timer
        from another corp
        then he is redirected back to dashboard
        """
        self.timer_3.visibility = Timer.VISIBILITY_CORPORATION
        self.timer_3.save()

        # login
        user_3 = create_test_user(self.character_3)
        user_3 = add_permission_to_user_by_name("structuretimers.create_timer", user_3)
        user_3 = add_permission_to_user_by_name("structuretimers.manage_timer", user_3)
        self.app.set_user(user_3)

        # user tries to access page for edit directly
        response = self.app.get(reverse("structuretimers:edit", args=[self.timer_3.pk]))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("authentication:dashboard"))

    def test_manager_tries_to_edit_opsec_timer_of_others(self):
        """
        given a user has permission to create and manage timers
        when trying to access page for timer edit of a opsec timer
        then he is redirected back to dashboard
        """
        self.timer_3.is_opsec = True
        self.timer_3.save()

        # login
        user_3 = create_test_user(self.character_3)
        user_3 = add_permission_to_user_by_name("structuretimers.create_timer", user_3)
        user_3 = add_permission_to_user_by_name("structuretimers.manage_timer", user_3)
        self.app.set_user(user_3)

        # user tries to access page for edit directly
        response = self.app.get(reverse("structuretimers:edit", args=[self.timer_3.pk]))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("authentication:dashboard"))

    def test_delete_existing_timer_by_manager(self):
        """
        when user has manager permissions
        then he can delete an existing timer
        """

        # login
        user_3 = create_test_user(self.character_3)
        user_3 = add_permission_to_user_by_name("structuretimers.manage_timer", user_3)
        self.app.set_user(user_3)

        # user opens timerboard
        timerboard = self.app.get(reverse("structuretimers:timer_list"))
        self.assertEqual(timerboard.status_code, 200)

        # user clicks on "Delete Timer" for timer 2
        confirm_page = self.app.get(
            reverse("structuretimers:delete", args=[self.timer_2.pk])
        )
        self.assertEqual(confirm_page.status_code, 200)

        # user enters data and clicks create
        form = confirm_page.forms["confirm-delete-form"]
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structuretimers:timer_list"))
        self.assertFalse(Timer.objects.filter(pk=self.timer_2.pk).exists())

    def test_delete_own_timer(self):
        """
        given a user has created a timer
        when trying to delete that time
        then timer is deleted
        """

        # login
        self.app.set_user(self.user_2)

        # user opens timerboard
        timerboard = self.app.get(reverse("structuretimers:timer_list"))
        self.assertEqual(timerboard.status_code, 200)

        # user clicks on "Delete Timer" for timer 2
        confirm_page = self.app.get(
            reverse("structuretimers:delete", args=[self.timer_2.pk])
        )
        self.assertEqual(confirm_page.status_code, 200)

        # user enters data and clicks create
        form = confirm_page.forms["confirm-delete-form"]
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structuretimers:timer_list"))
        self.assertFalse(Timer.objects.filter(pk=self.timer_2.pk).exists())

    def test_delete_timer_without_permission(self):
        """
        given a user does not have manager permissions
        when trying to access page to delete timer of another user
        then he is redirected back to dashboard
        """

        # login
        user_3 = create_test_user(self.character_3)
        user_3 = add_permission_to_user_by_name("structuretimers.create_timer", user_3)
        self.app.set_user(user_3)

        # user tries to access page for edit directly
        response = self.app.get(
            reverse("structuretimers:delete", args=[self.timer_2.pk])
        )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("authentication:dashboard"))


"""
@patch("structuretimers.models.sleep", new=lambda x: x)
@patch("structuretimers.models.dhooks_lite.Webhook.execute")
class TestSendNotifications(LoadTestDataMixin, TestCase):
    def setUp(self) -> None:
        self.webhook = DiscordWebhook.objects.create(
            name="Dummy", url="http://www.example.com"
        )
        self.rule = NotificationRule.objects.create(minutes=NotificationRule.MINUTES_0)
        self.rule.webhooks.add(self.webhook)

    def test_normal(self, mock_execute):
        Timer.objects.create(
            structure_name="Test_1",
            eve_solar_system=self.system_abune,
            structure_type=self.type_raitaru,
            date=now() + timedelta(seconds=2),
        )
        sleep(3)
        self.assertEqual(mock_execute.call_count, 1)
"""


@override_settings(CELERY_ALWAYS_EAGER=True)
@patch("structuretimers.models.sleep", new=lambda x: x)
@patch("structuretimers.tasks.notify", spec=True)
@patch("structuretimers.models.dhooks_lite.Webhook.execute", spec=True)
class TestTestMessageToWebhook(LoadTestDataMixin, TestCase):
    def setUp(self) -> None:
        self.webhook = DiscordWebhook.objects.create(
            name="Dummy", url="http://www.example.com"
        )
        self.user = AuthUtils.create_user("John Doe")

    def test_without_user(self, mock_execute, mock_notify):
        send_test_message_to_webhook.delay(webhook_pk=self.webhook.pk)
        self.assertEqual(mock_execute.call_count, 1)
        self.assertFalse(mock_notify.called)

    def test_with_user(self, mock_execute, mock_notify):
        send_test_message_to_webhook.delay(
            webhook_pk=self.webhook.pk, user_pk=self.user.pk
        )
        self.assertEqual(mock_execute.call_count, 1)
        self.assertTrue(mock_notify.called)



from datetime import timedelta

from django.contrib.auth.models import User
from django.db import models
from django.utils.timezone import now

from .app_settings import STRUCTURETIMERS_TIMERS_OBSOLETE_AFTER_DAYS


class NotificationRuleQuerySet(models.QuerySet):
    def conforms_with_timer(self, timer: object) -> models.QuerySet:
        """returns new queryset based on current queryset, which only contains notification rules that conforms with the given timer"""
        matching_rule_pks = list()
        for notification_rule in self:
            if notification_rule.is_matching_timer(timer):
                matching_rule_pks.append(notification_rule.pk)

        return self.filter(pk__in=matching_rule_pks)


class NotificationRuleManager(models.Manager):
    def get_queryset(self):
        return NotificationRuleQuerySet(self.model, using=self._db)


class TimerQuerySet(models.QuerySet):
    def conforms_with_notification_rule(
        self, notification_rule: object
    ) -> models.QuerySet:
        """returns new queryset based on current queryset, which only contains timers that conform with the given notification rule"""
        matching_timer_pks = list()
        for timer in self:
            if notification_rule.is_matching_timer(timer):
                matching_timer_pks.append(timer.pk)

        return self.filter(pk__in=matching_timer_pks)

    def visible_to_user(self, user: User) -> models.QuerySet:
        """returns updated queryset of all timers visible to the given user"""
        user_characters_qs = user.character_ownerships.select_related(
            "character_ownerships__character"
        ).values("character__corporation_id", "character__alliance_id")
        user_corporation_ids = {
            x["character__corporation_id"] for x in user_characters_qs
        }
        user_alliance_ids = {x["character__alliance_id"] for x in user_characters_qs}
        timers_qs = self.select_related(
            "structure_type", "eve_corporation", "eve_alliance"
        )
        if not user.has_perm("structuretimers.opsec_access"):
            timers_qs = timers_qs.exclude(is_opsec=True)

        timers_qs = (
            timers_qs.filter(visibility=self.model.VISIBILITY_UNRESTRICTED)
            | timers_qs.filter(user=user)
            | timers_qs.filter(
                visibility=self.model.VISIBILITY_CORPORATION,
                eve_corporation__corporation_id__in=user_corporation_ids,
            )
            | timers_qs.filter(
                visibility=self.model.VISIBILITY_ALLIANCE,
                eve_alliance__alliance_id__in=user_alliance_ids,
            )
        )
        return timers_qs


class TimerManager(models.Manager):
    def get_queryset(self):
        return TimerQuerySet(self.model, using=self._db)

    def delete_obsolete(self) -> int:
        """delete all timers that are considered obsolete"""
        if STRUCTURETIMERS_TIMERS_OBSOLETE_AFTER_DAYS:
            deadline = now() - timedelta(
                days=STRUCTURETIMERS_TIMERS_OBSOLETE_AFTER_DAYS
            )
            _, details = self.filter(date__lt=deadline).delete()
            key = f"{self.model._meta.app_label}.{self.model.__name__}"
            deleted_count = details[key]
            return deleted_count
        else:
            return 0



import socket
from datetime import datetime, timedelta
import json
import logging
import os
import re
from typing import Any
from urllib.parse import urljoin

from pytz import timezone

from django.apps import apps
from django.conf import settings
from django.contrib.messages.constants import DEBUG, ERROR, SUCCESS, WARNING, INFO
from django.contrib import messages
from django.db import models
from django.test import TestCase
from django.urls import reverse
from django.utils.functional import lazy
from django.utils.html import format_html, mark_safe
from django.utils.translation import gettext_lazy as _


# Format for output of datetime for this app
DATETIME_FORMAT = "%Y-%m-%d %H:%M"

format_html_lazy = lazy(format_html, str)


class LoggerAddTag(logging.LoggerAdapter):
    """add custom tag to a logger"""

    def __init__(self, my_logger, prefix):
        super(LoggerAddTag, self).__init__(my_logger, {})
        self.prefix = prefix

    def process(self, msg, kwargs):
        return "[%s] %s" % (self.prefix, msg), kwargs


logger = LoggerAddTag(logging.getLogger(__name__), __package__)


def get_swagger_spec_path() -> str:
    """returns the path to the current swagger spec file"""
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "swagger.json")


def make_logger_prefix(tag: str):
    """creates a function to add logger prefix, which returns tag when used empty"""
    return lambda text="": "{}{}".format(tag, (": " + text) if text else "")


class messages_plus:
    """Pendant to Django messages adding level icons and HTML support

    Careful: Use with safe strings only
    """

    _glyph_map = {
        DEBUG: "eye-open",
        INFO: "info-sign",
        SUCCESS: "ok-sign",
        WARNING: "exclamation-sign",
        ERROR: "alert",
    }

    @classmethod
    def _add_messages_icon(cls, level: int, message: str) -> str:
        """Adds an level based icon to standard Django messages"""
        if level not in cls._glyph_map:
            raise ValueError("glyph for level not defined")
        else:
            glyph = cls._glyph_map[level]

        return format_html(
            '<span class="glyphicon glyphicon-{}" '
            'aria-hidden="true"></span>&nbsp;&nbsp;{}',
            glyph,
            message,
        )

    @classmethod
    def debug(
        cls,
        request: object,
        message: str,
        extra_tags: str = "",
        fail_silently: bool = False,
    ):
        messages.debug(
            request, cls._add_messages_icon(DEBUG, message), extra_tags, fail_silently
        )

    @classmethod
    def info(
        cls,
        request: object,
        message: str,
        extra_tags: str = "",
        fail_silently: bool = False,
    ):
        messages.info(
            request, cls._add_messages_icon(INFO, message), extra_tags, fail_silently
        )

    @classmethod
    def success(
        cls,
        request: object,
        message: str,
        extra_tags: str = "",
        fail_silently: bool = False,
    ):
        messages.success(
            request, cls._add_messages_icon(SUCCESS, message), extra_tags, fail_silently
        )

    @classmethod
    def warning(
        cls,
        request: object,
        message: str,
        extra_tags: str = "",
        fail_silently: bool = False,
    ):
        messages.warning(
            request, cls._add_messages_icon(WARNING, message), extra_tags, fail_silently
        )

    @classmethod
    def error(
        cls,
        request: object,
        message: str,
        extra_tags: str = "",
        fail_silently: bool = False,
    ):
        messages.error(
            request, cls._add_messages_icon(ERROR, message), extra_tags, fail_silently
        )


def chunks(lst, size):
    """Yield successive sized chunks from lst."""
    for i in range(0, len(lst), size):
        yield lst[i : i + size]


def clean_setting(
    name: str,
    default_value: object,
    min_value: int = None,
    max_value: int = None,
    required_type: type = None,
    choices: list = None,
) -> Any:
    """cleans the input for a custom setting

    Will use `default_value` if settings does not exit or has the wrong type
    or is outside define boundaries (for int only)

    Need to define `required_type` if `default_value` is `None`

    Will assume `min_value` of 0 for int (can be overriden)

    `None` allowed as value

    Returns cleaned value for setting
    """
    if default_value is None and not required_type:
        raise ValueError("You must specify a required_type for None defaults")

    if not required_type:
        required_type = type(default_value)

    if min_value is None and required_type == int:
        min_value = 0

    if not hasattr(settings, name):
        cleaned_value = default_value
    else:
        dirty_value = getattr(settings, name)
        if dirty_value is None or (
            isinstance(dirty_value, required_type)
            and (min_value is None or dirty_value >= min_value)
            and (max_value is None or dirty_value <= max_value)
            and (choices is None or dirty_value in choices)
        ):
            cleaned_value = dirty_value
        else:
            logger.warn(
                "You setting for {} it not valid. Please correct it. "
                "Using default for now: {}".format(name, default_value)
            )
            cleaned_value = default_value
    return cleaned_value


def set_test_logger(logger_name: str, name: str) -> object:
    """set logger for current test module

    Args:
    - logger: current logger object
    - name: name of current module, e.g. __file__

    Returns:
    - amended logger
    """

    # reconfigure logger so we get logging from tested module
    f_format = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(module)s:%(funcName)s - %(message)s"
    )
    f_handler = logging.FileHandler("{}.log".format(os.path.splitext(name)[0]), "w+")
    f_handler.setFormatter(f_format)
    my_logger = logging.getLogger(logger_name)
    my_logger.level = logging.DEBUG
    my_logger.addHandler(f_handler)
    my_logger.propagate = False
    return my_logger


def timeuntil_str(duration: timedelta) -> str:
    """return the duration as nicely formatted string.
    Or empty string if duration is negative.

    Format: '[[[999y] [99m]] 99d] 99h 99m 99s'
    """
    seconds = int(duration.total_seconds())
    if seconds > 0:
        periods = [
            # Translators: Abbreviation for years
            (_("y"), 60 * 60 * 24 * 365, False),
            # Translators: Abbreviation for months
            (_("mt"), 60 * 60 * 24 * 30, False),
            # Translators: Abbreviation for days
            (_("d"), 60 * 60 * 24, False),
            # Translators: Abbreviation for hours
            (_("h"), 60 * 60, True),
            # Translators: Abbreviation for months
            (_("m"), 60, True),
            # Translators: Abbreviation for seconds
            (_("s"), 1, True),
        ]
        strings = list()
        for period_name, period_seconds, period_static in periods:
            if seconds >= period_seconds or period_static:
                period_value, seconds = divmod(seconds, period_seconds)
                strings.append("{}{}".format(period_value, period_name))

        result = " ".join(strings)
    else:
        result = ""

    return result


class SocketAccessError(Exception):
    pass


class NoSocketsTestCase(TestCase):
    """Variation of TestCase class that prevents any use of sockets"""

    @classmethod
    def setUpClass(cls):
        cls.socket_original = socket.socket
        socket.socket = cls.guard
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        socket.socket = cls.socket_original
        return super().tearDownClass()

    @staticmethod
    def guard(*args, **kwargs):
        raise SocketAccessError("Attempted to access network")


def app_labels() -> set:
    """returns set of all current app labels"""
    return {x for x in apps.app_configs.keys()}


def add_no_wrap_html(text: str) -> str:
    """add no-wrap HTML to text"""
    return format_html('<span style="white-space: nowrap;">{}</span>', mark_safe(text))


def yesno_str(value: bool) -> str:
    """returns yes/no for boolean as string and with localization"""
    return _("yes") if value is True else _("no")


def get_site_base_url() -> str:
    """return base URL for this site"""
    try:
        match = re.match(r"(.+)\/sso\/callback", settings.ESI_SSO_CALLBACK_URL)
        if match:
            return match.group(1)
    except AttributeError:
        pass

    return ""


def get_absolute_url(url_name: str) -> str:
    """Returns absolute URL for the given URL name."""
    return urljoin(get_site_base_url(), reverse(url_name))


def dt_eveformat(dt: object) -> str:
    """converts a datetime to a string in eve format
    e.g. '2019-06-25T19:04:44'
    """
    dt2 = datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)
    return dt2.isoformat()


def add_bs_label_html(text: str, label: str) -> str:
    """create Bootstrap label and return HTML"""
    return format_html('<div class="label label-{}">{}</div>', label, text)


def create_link_html(url: str, label: str, new_window: bool = True) -> str:
    """create html link and return HTML"""
    return format_html(
        '<a href="{}"{}>{}</a>',
        url,
        mark_safe(' target="_blank"') if new_window else "",
        label,
    )


def create_bs_glyph_html(glyph_name: str) -> str:
    return format_html(
        '<span class="glyphicon glyphicon-{}"></span>', glyph_name.lower()
    )


def create_bs_glyph_2_html(glyph_name, tooltip_text=None, color="initial"):
    if tooltip_text:
        tooltip_html = mark_safe(
            'aria-hidden="true" data-toggle="tooltip" data-placement="top" '
            'title="{}"'.format(tooltip_text)
        )
    else:
        tooltip_html = ""
    return format_html(
        '<span class="glyphicon glyphicon-{}"'
        ' style="color:{};"{}></span>'.format(glyph_name.lower(), color, tooltip_html)
    )


def create_bs_button_html(
    url: str, glyph_name: str, button_type: str, disabled: bool = False
) -> str:
    """create BS botton and return HTML"""
    return format_html(
        '<a href="{}" class="btn btn-{}"{}>{}</a>',
        url,
        button_type,
        mark_safe(' disabled="disabled"') if disabled else "",
        create_bs_glyph_html(glyph_name),
    )


def create_fa_button_html(
    url: str,
    fa_code: str,
    button_type: str,
    tooltip: str = None,
    disabled: bool = False,
) -> str:
    """create BS botton and return HTML"""
    return format_html(
        '<a href="{}" class="btn btn-{}"{}>{}{}</a>',
        url,
        button_type,
        mark_safe(f' title="{tooltip}"') if tooltip else "",
        mark_safe(' disabled="disabled"') if disabled else "",
        mark_safe(f'<i class="{fa_code}"></i>'),
    )


class JSONDateTimeDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs) -> None:
        json.JSONDecoder.__init__(
            self, object_hook=self.dict_to_object, *args, **kwargs
        )

    def dict_to_object(self, dct: dict) -> object:
        if "__type__" not in dct:
            return dct

        type_str = dct.pop("__type__")
        zone, _ = dct.pop("tz")
        dct["tzinfo"] = timezone(zone)
        try:
            dateobj = datetime(**dct)
            return dateobj
        except (ValueError, TypeError):
            dct["__type__"] = type_str
            return dct


class JSONDateTimeEncoder(json.JSONEncoder):
    """Instead of letting the default encoder convert datetime to string,
    convert datetime objects into a dict, which can be decoded by the
    JSONDateTimeDecoder
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, datetime):
            return {
                "__type__": "datetime",
                "year": o.year,
                "month": o.month,
                "day": o.day,
                "hour": o.hour,
                "minute": o.minute,
                "second": o.second,
                "microsecond": o.microsecond,
                "tz": (o.tzinfo.tzname(o), o.utcoffset().total_seconds()),
            }
        else:
            return json.JSONEncoder.default(self, o)


def generate_invalid_pk(MyModel: models.Model) -> int:
    """return an invalid PK for the given Django model"""
    return MyModel.objects.aggregate(models.Max("pk"))["pk__max"] + 1
